#!/usr/bin/env python3

## Create an embedding matrix from .vec files (fasttext format)
# Copyright (C) <2018-2022>  <Agence Data Services, DSI Pôle Emploi>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# Ex: python 0_get_embedding_dict.py  --filename cc.fr.300.vec


import os
import pickle
import logging
import argparse
import numpy as np
import pandas as pd

from {{package_name}} import utils

# Get logger
logger = logging.getLogger('{{package_name}}.0_get_embedding_dict')


def main(filename: str, encoding: str = '{{default_encoding}}') -> None:
    '''Creates an embedding matrix from .vec files (fasttext format)

    Those files can be downloaded here (text files): https://fasttext.cc/docs/en/crawl-vectors.html

    Args:
        filename (str): A .vec file name (actually a path relative to {{package_name}}-data)
    Kwargs:
        encoding (str): Encoding to use with the .vec file
    Raises:
        FileNotFoundError: If the file does not exist in {{package_name}}-data
    '''
    logger.info("Creating an embedding matrix ...")

    # Manage path
    data_path = utils.get_data_path()
    embedding_path = os.path.join(data_path, filename)
    if not os.path.exists(embedding_path):
        raise FileNotFoundError(f"The file {embedding_path} does not exist")

    # Get embedding indexes
    def get_coefs(word, *arr):
        return word, np.asarray(arr, dtype='float32')
    embedding_indexes = dict(get_coefs(*o.strip().split(" ")) for o in open(embedding_path, encoding=encoding))

    # Save
    file_path = os.path.join(data_path, '.'.join(filename.split('.')[:-1]) + '.pkl')
    with open(file_path, 'wb') as f:
        pickle.dump(embedding_indexes, f, pickle.HIGHEST_PROTOCOL)


if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument('-f', '--filename', required=True, help='A .vec file name (actually a path relative to {{package_name}}-data)')
    parser.add_argument('--encoding', default="utf-8", help="Encoding to use with the .vec file.")
    args = parser.parse_args()
    main(filename=args.filename, encoding=args.encoding)



# Copyright (C) <2018-2022>  <Agence Data Services, DSI Pôle Emploi>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


import os
from setuptools import setup

# Get package version
with open(os.path.join(os.path.dirname(os.path.abspath(__file__)), 'version.txt'), 'r') as version_file:
    version = version_file.read().strip()

version = os.getenv('VERSION') or f"{version}-local"
# Setup
setup(
    name="{{package_name}}",
    version=version,
    packages=["{{package_name}}", "{{package_name}}.preprocessing", "{{package_name}}.models_training", "{{package_name}}.monitoring"],
    url="",
    license="",
    author="Agence Data Services PE Aix / Nantes",
    author_email="",
    description="",
    include_package_data=True,
    package_data={'': ['{{package_name}}/configs/**']},
    install_requires=[
        'numpy==1.19.5',
        'pandas==1.3.5',
        'scikit_learn>=0.24.2,<0.25',
        'matplotlib>=3.0.3,<3.4',
        'seaborn>=0.9.0,<0.12',
        'opencv-python-headless==4.5.5.62',
        'dill>=0.3.2,<0.3.4',
        'mlflow>=1.11.0,<1.12.2',
        'tensorflow==2.6.2',
        'pycocotools==2.0.4',
        'tqdm==4.62.2',  #https://github.com/tqdm/tqdm/issues/780
        'protobuf==3.20.1',  #https://developers.google.com/protocol-buffers/docs/news/2022-05-06#python-updates
    ],
    extras_require={
        "detectron": ["torch==1.8.1+cpu", "detectron2==0.6+cpu", "torchvision==0.9.1+cpu"],  # If GPU with cuda 11.1 : replace +cpu by +cu111
    }
    # pip install {{package_name}} || pip install {{package_name}}[detectron]
)



#!/usr/bin/env python3

## Utils - tools-functions for deep_learning keras models
# Copyright (C) <2018-2022>  <Agence Data Services, DSI Pôle Emploi>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# Functions :
# - f1 -> f1 score, to use as custom metrics
# - f1_loss -> f1 loss, to use as custom loss


import logging
from functools import partial
from typing import Callable, Any

import tensorflow as tf
from tensorflow.keras import backend as K
from tensorflow.keras import initializers, regularizers, constraints, activations
from tensorflow.keras.layers import Layer
from tensorflow.keras.activations import softmax

from {{package_name}} import utils

# Get logger
logger = logging.getLogger(__name__)



def recall(y_true, y_pred) -> float:
    '''Recall to use as a custom metrics

    Args:
        y_true: Ground truth values
        y_pred: The predicted values
    Returns:
        float: metric
    '''
    y_pred = K.round(y_pred)
    y_true = K.cast(y_true, 'float32') # Fix : TypeError: Input 'y' of 'Mul' Op has type float32 that does not match type int32 of argument 'x'.

    ground_positives = K.sum(y_true, axis=0) + K.epsilon()  # We add an epsilon -> manage the case where a class is absent in the batch

    tp = K.sum(K.cast(y_true * y_pred, 'float'), axis=0)
    fn = K.sum(K.cast(y_true * (1 - y_pred), 'float'), axis=0)

    recall = tp / (tp + fn + K.epsilon())
    recall = tf.where(tf.math.is_nan(recall), tf.zeros_like(recall), recall)

    weighted_recall = recall * ground_positives / K.sum(ground_positives)
    weighted_recall = K.sum(weighted_recall)

    return weighted_recall


def precision(y_true, y_pred) -> float:
    '''Precision, to use as custom metrics

    Args:
        y_true: Ground truth values
        y_pred: The predicted values
    Returns:
        float: metric
    '''
    y_pred = K.round(y_pred)
    y_true = K.cast(y_true, 'float32') # Fix : TypeError: Input 'y' of 'Mul' Op has type float32 that does not match type int32 of argument 'x'.

    ground_positives = K.sum(y_true, axis=0) + K.epsilon()  # We add an epsilon -> manage the case where a class is absent in the batch

    tp = K.sum(K.cast(y_true * y_pred, 'float'), axis=0)
    fp = K.sum(K.cast((1 - y_true) * y_pred, 'float'), axis=0)

    precision = tp / (tp + fp + K.epsilon())
    precision = tf.where(tf.math.is_nan(precision), tf.zeros_like(precision), precision)

    weighted_precision = precision * ground_positives / K.sum(ground_positives)
    weighted_precision = K.sum(weighted_precision)

    return weighted_precision


def f1(y_true, y_pred) -> float:
    '''f1 score, to use as custom metrics

    - /!\\ To use with a big batch size /!\\ -

    From:
        https://www.kaggle.com/rejpalcz/best-loss-function-for-f1-score-metric
        https://stackoverflow.com/questions/59963911/how-to-write-a-custom-f1-loss-function-with-weighted-average-for-keras

    Args:
        y_true: Ground truth values
        y_pred: The predicted values
    Returns:
        float: metric
    '''
    # Round pred to 0 & 1
    y_pred = K.round(y_pred)
    y_true = K.cast(y_true, 'float32') # Fix : TypeError: Input 'y' of 'Mul' Op has type float32 that does not match type int32 of argument 'x'.

    ground_positives = K.sum(y_true, axis=0) + K.epsilon()  # We add an epsilon -> manage the case where a class is absent in the batch

    y_pred = K.round(y_pred)
    tp = K.sum(K.cast(y_true * y_pred, 'float'), axis=0)
    tn = K.sum(K.cast((1 - y_true) * (1 - y_pred), 'float'), axis=0)
    fp = K.sum(K.cast((1 - y_true) * y_pred, 'float'), axis=0)
    fn = K.sum(K.cast(y_true * (1 - y_pred), 'float'), axis=0)

    p = tp / (tp + fp + K.epsilon())
    r = tp / (tp + fn + K.epsilon())

    f1 = 2 * p * r / (p + r + K.epsilon())
    f1 = tf.where(tf.math.is_nan(f1), tf.zeros_like(f1), f1)

    weighted_f1 = f1 * ground_positives / K.sum(ground_positives)
    weighted_f1 = K.sum(weighted_f1)

    return weighted_f1


def f1_loss(y_true, y_pred) -> float:
    '''f1 loss, to use as custom loss

    - /!\\ To use with a big batch size /!\\ -

    From:
        https://www.kaggle.com/rejpalcz/best-loss-function-for-f1-score-metric
        https://stackoverflow.com/questions/59963911/how-to-write-a-custom-f1-loss-function-with-weighted-average-for-keras

    Args:
        y_true: Ground truth values
        y_pred: The predicted values
    Returns:
        float: metric
    '''
    # TODO : Find a mean of rounding y_pred
    # TODO : Problem : models will quickly converge on probabilities 1.0 & 0.0 to optimize this loss....
    # We can't round here :(
    # Please make sure that all of your ops have a gradient defined (i.e. are differentiable).
    # Common ops without gradient: K.argmax, K.round, K.eval.
    y_true = K.cast(y_true, 'float32') # Fix : TypeError: Input 'y' of 'Mul' Op has type float32 that does not match type int32 of argument 'x'.

    ground_positives = K.sum(y_true, axis=0) + K.epsilon()  # We add an epsilon -> manage the case where a class is absent in the batch

    tp = K.sum(K.cast(y_true * y_pred, 'float'), axis=0)
    tn = K.sum(K.cast((1 - y_true) * (1 - y_pred), 'float'), axis=0)
    fp = K.sum(K.cast((1 - y_true) * y_pred, 'float'), axis=0)
    fn = K.sum(K.cast(y_true * (1 - y_pred), 'float'), axis=0)

    p = tp / (tp + fp + K.epsilon())
    r = tp / (tp + fn + K.epsilon())

    f1 = 2 * p * r / (p + r + K.epsilon())
    f1 = tf.where(tf.math.is_nan(f1), tf.zeros_like(f1), f1)

    weighted_f1 = f1 * ground_positives / K.sum(ground_positives)
    weighted_f1 = K.sum(weighted_f1)

    return 1 - weighted_f1


def fb_loss(b: float, y_true, y_pred) -> float:
    '''fB loss, to use as custom loss

    - /!\\ To use with a big batch size /!\\ -

    From:
        https://www.kaggle.com/rejpalcz/best-loss-function-for-f1-score-metric
        https://stackoverflow.com/questions/59963911/how-to-write-a-custom-f1-loss-function-with-weighted-average-for-keras

    Args:
        b (float): importance recall in the calculation of the fB score
        y_true: Ground truth values
        y_pred: The predicted values
    Returns:
        float: metric
    '''
    # TODO : Find a mean of rounding y_pred
    # TODO : Problem : models will quickly converge on probabilities 1.0 & 0.0 to optimize this loss....
    # We can't round here :(
    # Please make sure that all of your ops have a gradient defined (i.e. are differentiable).
    # Common ops without gradient: K.argmax, K.round, K.eval.
    y_true = K.cast(y_true, 'float32') # Fix : TypeError: Input 'y' of 'Mul' Op has type float32 that does not match type int32 of argument 'x'.

    ground_positives = K.sum(y_true, axis=0) + K.epsilon()  # We add an epsilon -> manage the case where a class is absent in the batch

    tp = K.sum(K.cast(y_true * y_pred, 'float'), axis=0)
    tn = K.sum(K.cast((1 - y_true) * (1 - y_pred), 'float'), axis=0)
    fp = K.sum(K.cast((1 - y_true) * y_pred, 'float'), axis=0)
    fn = K.sum(K.cast(y_true * (1 - y_pred), 'float'), axis=0)

    p = tp / (tp + fp + K.epsilon())
    r = tp / (tp + fn + K.epsilon())

    fb = (1 + b**2) * p * r / ((p * b**2) + r + K.epsilon())
    fb = tf.where(tf.math.is_nan(fb), tf.zeros_like(fb), fb)

    weighted_fb = fb * ground_positives / K.sum(ground_positives)
    weighted_fb = K.sum(weighted_fb)

    return 1 - weighted_fb


def get_fb_loss(b: float = 2.0) -> Callable:
    ''' Gets a fB-score loss

    Args:
        b (float): importance recall in the calculation of the fB score
    Returns:
        Callable: fb_loss
    '''
    # - /!\ Utilisation partial obligatoire pour pouvoir pickle des fonctions dynamiques ! /!\ -
    fn = partial(fb_loss, b)
    # FIX:  AttributeError: 'functools.partial' object has no attribute '__name__'
    fn.__name__ = 'fb_loss'  # type: ignore
    return fn


def weighted_binary_crossentropy(pos_weight: float, target, output) -> float:
    '''Weighted binary crossentropy between an output tensor
    and a target tensor. pos_weight is used as a multiplier
    for the positive targets.

    Combination of the following functions:
    * keras.losses.binary_crossentropy
    * keras.backend.tensorflow_backend.binary_crossentropy
    * tf.nn.weighted_cross_entropy_with_logits

    Args:
        pos_weight (float): poid classe positive, to be tuned
        target: Target tensor
        output: Output tensor
    Returns:
        float: metric
    '''
    target = K.cast(target, 'float32')
    output = K.cast(output, 'float32')
    # transform back to logits
    _epsilon = tf.convert_to_tensor(K.epsilon(), output.dtype.base_dtype)
    output = tf.clip_by_value(output, _epsilon, 1 - _epsilon)
    output = tf.math.log(output / (1 - output))
    # compute weighted loss
    loss = tf.nn.weighted_cross_entropy_with_logits(target, output, pos_weight=pos_weight)
    return tf.reduce_mean(loss, axis=-1)


def get_weighted_binary_crossentropy(pos_weight: float = 10.0) -> Callable:
    ''' Gets a "weighted binary crossentropy" loss
    From https://stats.stackexchange.com/questions/261128/neural-network-for-multi-label-classification-with-large-number-of-classes-outpu
    TO BE ADDED IN custom_objects : 'weighted_binary_crossentropy': utils_deep_keras.get_weighted_binary_crossentropy(pos_weight=...)

    Args:
        pos_weight (float): Weight of the positive class, to be tuned
    Returns:
        Callable: Weighted binary crossentropy loss
    '''
    # - /!\ Use of partial mandatory in order to be able to pickle dynamical functions ! /!\ -
    fn = partial(weighted_binary_crossentropy, pos_weight)
    # FIX:  AttributeError: 'functools.partial' object has no attribute '__name__'
    fn.__name__ = 'weighted_binary_crossentropy'  # type: ignore
    return fn


# ** EXPERIMENTAL **
# ** EXPERIMENTAL **
# ** EXPERIMENTAL **

# From Gaëlle JOUIS Thesis

class AttentionAverage(Layer):
    def __init__(self, attention_hops, **kwargs) -> None:
            self.attention_hops = attention_hops
            self.applied_axis = 1
            super(AttentionAverage, self).__init__()

    def get_config(self) -> Any:
        '''Gets the config'''
        config = super().get_config().copy()
        config.update({
            'attention_hops': self.attention_hops
        })
        return config

    def call(self, input) -> Any:
        return tf.divide(tf.reduce_sum(input, self.applied_axis), self.attention_hops)


class AttentionWithContext(Layer):
    '''Attention operation, with a context/query vector, for temporal data.
    Supports Masking.
    Follows the work of Yang et al. [https://www.cs.cmu.edu/~diyiy/docs/naacl16.pdf]
    "Hierarchical Attention Networks for Document Classification"
    by using a context vector to assist the attention
    # Input shape
        3D tensor with shape: `(samples, steps, features)`.
    # Output shape
        2D tensor with shape: `(samples, features)`.
    :param kwargs:
    Just put it on top of an RNN Layer (GRU/LSTM/SimpleRNN) with return_sequences=True.
    The dimensions are inferred based on the output shape of the RNN.
    Example:
        model.add(LSTM(64, return_sequences=True))
        model.add(AttentionWithContext())
    '''
    def __init__(self, W_regularizer=None, u_regularizer=None, b_regularizer=None,
                 W_constraint=None, u_constraint=None, b_constraint=None, bias=True,
                 return_attention=False, **kwargs):
        self.return_attention = return_attention
        self.bias = bias
        super(AttentionWithContext, self).__init__(**kwargs)

    def get_config(self) -> Any:
        config = super().get_config().copy()
        config.update({
            'return_attention': self.return_attention,
            'bias': self.bias
        })
        return config

    def build(self, input_shape) -> None:
        assert len(input_shape) == 3
        input_shape_list = input_shape.as_list()

        self.W = self.add_weight(shape=((input_shape_list[-1], input_shape_list[-1])),
                                 name='{}_W'.format(self.name))
        if self.bias:
            self.b = self.add_weight(shape=(input_shape_list[-1],),
                                     name='{}_b'.format(self.name))

        self.u = self.add_weight(shape=(input_shape_list[-1],),
                                 name='{}_u'.format(self.name))

        super(AttentionWithContext, self).build(input_shape.as_list())

    def compute_mask(self, input, input_mask=None) -> Any:
        # do not pass the mask to the next layers
        return None

    def call(self, x, mask=None) -> Any:
        uit = tf.tensordot(x, self.W, axes=1)
        if self.bias:
            uit += self.b
        uit = activations.tanh(uit)
        ait = tf.tensordot(uit, self.u, axes=1)
        a = activations.exponential(ait)
        # Apply mask after the exp. will be re-normalized next
        if mask is not None:
            # Cast the mask to floatX to avoid float64 upcasting in theano
            a *= tf.cast(mask, K.floatx())
        # In some cases especially in the early stages of training the sum may be almost zero
        # and this results in NaN's. A workaround is to add a very small positive number ε to the sum.
        # a /= K.cast(K.sum(a, axis=1, keepdims=True), K.floatx())
        a /= tf.cast(K.sum(a, axis=1, keepdims=True) + K.epsilon(), K.floatx())
        a = K.expand_dims(a)
        weighted_input = x * a
        result = K.sum(weighted_input, axis=1)
        if self.return_attention:
            return [result, a]
        return result

    def compute_output_shape(self, input_shape) -> Any:
        if self.return_attention:
            return [(input_shape[0], input_shape[-1]), (input_shape[0], input_shape[1])]
        else:
            return tf.TensorShape([input_shape[0].value, input_shape[-1].value])


def softmax_axis(x) -> float:
    return softmax(x, axis=1)

# ** EXPERIMENTAL **
# ** EXPERIMENTAL **
# ** EXPERIMENTAL **


# /!\ Important -> This dictionary defines the "custom" objets used in our models
# /!\ Important -> They are mandatory in order to serialize and save the model
# /!\ Important -> All customs objects must be added to it
# TODO: to be improved
custom_objects = {
    'f1': f1,
    'f1_loss': f1_loss,
    'recall': recall,
    'precision': precision,
    'AttentionWithContext': AttentionWithContext,
    'AttentionAverage': AttentionAverage,
    'softmax_axis': softmax_axis,
}


if __name__ == '__main__':
    logger.error("This script is not stand alone but belongs to a package that has to be imported.")



#!/usr/bin/env python3
# Copyright (C) <2018-2022>  <Agence Data Services, DSI Pôle Emploi>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

# Libs unittest
import unittest
from unittest.mock import Mock
from unittest.mock import patch


# Utils libs
import os
import json
import numpy as np
import pandas as pd
from {{package_name}} import utils
from {{package_name}}.preprocessing import column_preprocessors

# Disable logging
import logging
logging.disable(logging.CRITICAL)


class ColumnPreprocessorsTests(unittest.TestCase):
    '''Main class to test all functions in {{package_name}}.preprocessing.column_preprocessors'''


    def setUp(self):
        '''SetUp fonction'''
        # Change directory to script directory
        abspath = os.path.abspath(__file__)
        dname = os.path.dirname(abspath)
        os.chdir(dname)


    def test01_AutoLogTransform(self):
        '''Test of the class column_preprocessors.AutoLogTransform'''
        # Values to test
        df = pd.DataFrame({
            'skewed_1': [1, 2, 3, 4, 5, 6, 7, 1000, 100000, 10000000] * 100,
            'not_skewed': [1, 2, 3, 4, 5, 6, 7, 8, 9, 10] * 100,
            'skewed_2': [-10000000, -100000, -1000, 1, 2, 3, 4, 5, 6, 7] * 100, # For now, we accept negatives but returns nans !
        })
        df_copy = df.copy(deep=True)
        arr = df.to_numpy()
        arr_copy = arr.copy()
        transformed_arr = pd.DataFrame({
            'skewed_1': np.log([1, 2, 3, 4, 5, 6, 7, 1000, 100000, 10000000] * 100),
            'not_skewed': [1, 2, 3, 4, 5, 6, 7, 8, 9, 10] * 100,
            'skewed_2': np.log([-10000000, -100000, -1000, 1, 2, 3, 4, 5, 6, 7] * 100),
        }).to_numpy()

        new_df =  pd.DataFrame({
            'toto': [1, 2, 3, 4, 5, 6, 7, 8, 9, 10],
            'titi': [-10000000, -100000, -1000, 1, 2, 3, 4, 5, 6, 7],
            'tata': [5, 2, 4, 4, 5, 68, 7, 81, 9, -10],
        })
        new_df_copy = new_df.copy(deep=True)
        new_transformed_arr = pd.DataFrame({
            'toto': np.log([1, 2, 3, 4, 5, 6, 7, 8, 9, 10]),
            'titi': [-10000000, -100000, -1000, 1, 2, 3, 4, 5, 6, 7],
            'tata': np.log([5, 2, 4, 4, 5, 68, 7, 81, 9, -10]),
        }).to_numpy()

        # Nominal case
        transformer = column_preprocessors.AutoLogTransform()
        transformer.fit(df)
        self.assertTrue(hasattr(transformer, 'fitted_'))
        self.assertEqual(transformer.applicable_columns_index, [0, 2])
        np.testing.assert_array_equal(transformer.transform(df), transformed_arr)
        np.testing.assert_array_equal(transformer.transform(arr), transformed_arr)
        np.testing.assert_array_equal(transformer.transform(new_df), new_transformed_arr)
        pd.testing.assert_frame_equal(df, df_copy)
        pd.testing.assert_frame_equal(new_df, new_df_copy)
        np.testing.assert_array_equal(arr, arr_copy)

        # np array case
        transformer = column_preprocessors.AutoLogTransform()
        transformer.fit(arr)
        self.assertEqual(transformer.applicable_columns_index, [0, 2])
        np.testing.assert_array_equal(transformer.transform(arr), transformed_arr)
        np.testing.assert_array_equal(transformer.transform(arr), transformed_arr)
        np.testing.assert_array_equal(transformer.transform(new_df), new_transformed_arr)
        np.testing.assert_array_equal(arr, arr_copy)
        pd.testing.assert_frame_equal(new_df, new_df_copy)
        np.testing.assert_array_equal(arr, arr_copy)

        # Nominal case - fit_transform
        transformer = column_preprocessors.AutoLogTransform(min_skewness=1, min_amplitude=1000)
        self.assertEqual(transformer.applicable_columns_index, None)
        self.assertEqual(transformer.min_skewness, 1)
        self.assertEqual(transformer.min_amplitude, 1000)
        np.testing.assert_array_equal(transformer.fit_transform(df), transformed_arr)
        pd.testing.assert_frame_equal(df, df_copy)

    def test02_ThresholdingTransform(self):
        '''Test of the class column_preprocessors.ThresholdingTransform'''
        # Values to test
        df = pd.DataFrame({
            'x': [1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11],
            'y': [1, 2, 3, 4, 5, 6, 7, 1000, 100000, 10000000, 100000000],
            'z': [1, 2, 3, 4, 5, 6, 156, 5648, 10000, 10001, 10002],
        })
        df_copy = df.copy(deep=True)
        arr = df.to_numpy()
        arr_copy = arr.copy()
        transformed_arr = pd.DataFrame({
            'x': [2, 2, 3, 4, 5, 6, 6, 6, 6, 6, 6],
            'y': [1.5, 2, 3, 4, 5, 6, 7, 100, 100, 100, 100],
            'z': [1.5, 2, 3, 4, 5, 6, 156, 5648, 10000, 10001, 10001.5],
        }).to_numpy()

        new_df =  pd.DataFrame({
            'toto': [-10, -20, -5, 4, 5, 6, 7, 8, 9, 10, 11],
            'titi': [-10000000, -100000, -1000, 1, 2, 3, 4, 5, 6, 1000000, 100000000],
            'tata': [1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11],
        })
        new_df_copy = new_df.copy(deep=True)
        new_transformed_arr = pd.DataFrame({
            'toto': [2, 2, 2, 4, 5, 6, 6, 6, 6, 6, 6],
            'titi': [1.5, 1.5, 1.5, 1.5, 2, 3, 4, 5, 6, 100, 100],
            'tata': [1.5, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11],
        }).to_numpy()

        # Nominal case
        transformer = column_preprocessors.ThresholdingTransform(thresholds=[(2, 6), (None, 100), (None, None)])
        transformer.fit(df)
        self.assertTrue(hasattr(transformer, 'fitted_'))
        self.assertEqual(len(transformer.fitted_thresholds), 3)
        np.testing.assert_array_equal(transformer.transform(df), transformed_arr)
        np.testing.assert_array_equal(transformer.transform(arr), transformed_arr)
        np.testing.assert_array_equal(transformer.transform(new_df), new_transformed_arr)
        pd.testing.assert_frame_equal(df, df_copy)
        pd.testing.assert_frame_equal(new_df, new_df_copy)
        np.testing.assert_array_equal(arr, arr_copy)

        # Quantiles !=
        transformed_arr_quantiles_diff = pd.DataFrame({
            'x': [2, 2, 3, 4, 5, 6, 6, 6, 6, 6, 6],
            'y': [2, 2, 3, 4, 5, 6, 7, 100, 100, 100, 100],
            'z': [2, 2, 3, 4, 5, 6, 156, 5648, 10000, 10001, 10001],
        }).to_numpy()
        new_transformed_arr_quantiles_diff = pd.DataFrame({
            'toto': [2, 2, 2, 4, 5, 6, 6, 6, 6, 6, 6],
            'titi': [2, 2, 2, 2, 2, 3, 4, 5, 6, 100, 100],
            'tata': [2, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11],
        }).to_numpy()
        transformer = column_preprocessors.ThresholdingTransform(thresholds=[(2, 6), (None, 100), (None, None)], quantiles=(0.10, 0.90))
        transformer.fit(arr) # Fit on a numpy array to test while at it
        self.assertTrue(hasattr(transformer, 'fitted_'))
        self.assertEqual(len(transformer.fitted_thresholds), 3)
        np.testing.assert_array_equal(transformer.transform(df), transformed_arr_quantiles_diff)
        np.testing.assert_array_equal(transformer.transform(new_df), new_transformed_arr_quantiles_diff)
        pd.testing.assert_frame_equal(df, df_copy)
        pd.testing.assert_frame_equal(new_df, new_df_copy)

        # Nominal case - fit_transform
        transformer = column_preprocessors.ThresholdingTransform(thresholds=[(2, 6), (None, 100), (None, None)])
        self.assertFalse(hasattr(transformer, 'fitted_'))
        self.assertEqual(len(transformer.fitted_thresholds), 0)
        np.testing.assert_array_equal(transformer.fit_transform(df), transformed_arr)
        pd.testing.assert_frame_equal(df, df_copy)

        # Check errors
        with self.assertRaises(ValueError):
            column_preprocessors.ThresholdingTransform(thresholds=[(2, 6)], quantiles=(-1, 0.98))
        with self.assertRaises(ValueError):
            column_preprocessors.ThresholdingTransform(thresholds=[(2, 6)], quantiles=(5, 0.98))
        with self.assertRaises(ValueError):
            column_preprocessors.ThresholdingTransform(thresholds=[(2, 6)], quantiles=(0.05, -1))
        with self.assertRaises(ValueError):
            column_preprocessors.ThresholdingTransform(thresholds=[(2, 6)], quantiles=(0.05, 5))
        with self.assertRaises(ValueError):
            column_preprocessors.ThresholdingTransform(thresholds=[(2, 6)], quantiles=(0.95, 0.05))
        transformer = column_preprocessors.ThresholdingTransform(thresholds=[(2, 6), (None, 100), (None, None)])
        transformer.fit(df)

    def test03_AutoBinner(self):
        '''Test of the class column_preprocessors.AutoBinner'''
        # Values to test
        # 4030 values
        # threshold 0.05 -> 201.5
        # strategy = trehsold -> unique vals > 201 OK
        #
        df = pd.DataFrame({
            'x': ['a', 'b', 'b', 'c'] * 1000 + ['d', 'e', 'f'] * 10,
            'y': ['a', 'b', 'b', 'c'] * 1006 + ['d', 'e', 'f'] * 2,
            'z': ['a', 'b', 'b', 'c'] * 930 + ['d'] * 200 + ['e', 'f'] * 55,
        })
        df_copy = df.copy(deep=True)
        arr = df.to_numpy()
        arr_copy = arr.copy()
        transformed_arr = pd.DataFrame({
            'x': ['a', 'b', 'b', 'c'] * 1000 + ['other_', 'other_', 'other_'] * 10,
            'y': ['a', 'b', 'b', 'c'] * 1006 + ['other_', 'other_', 'other_'] * 2,
            'z': ['a', 'b', 'b', 'c'] * 930 + ['d'] * 200 + ['other_', 'other_'] * 55,
        }).to_numpy()

        new_df =  pd.DataFrame({
            'toto': ['a', 'g', 'b', 'c'] * 1000 + ['d', 'e', 'f'] * 10,
            'titi': ['d', 'd', 'e', 'f'] * 1006 + ['a', 'b', 'c'] * 2,
            'tata': ['i', 'j', 'b', 'c'] * 930 + ['d'] * 200 + ['e', 'f'] * 55,
        })
        new_df_copy = new_df.copy(deep=True)
        new_transformed_arr = pd.DataFrame({
            'toto': ['a', 'other_', 'b', 'c'] * 1000 + ['other_', 'other_', 'other_'] * 10,
            'titi': ['other_', 'other_', 'other_', 'other_'] * 1006 + ['a', 'b', 'c'] * 2,
            'tata': ['other_', 'other_', 'b', 'c'] * 930 + ['d'] * 200 + ['other_', 'other_'] * 55,
        }).to_numpy()

        # Nominal case
        transformer = column_preprocessors.AutoBinner(strategy="auto", min_cat_count=3, threshold=0.05)
        transformer.fit(df)
        self.assertTrue(hasattr(transformer, 'fitted_'))
        self.assertEqual(transformer.input_length, 3)
        np.testing.assert_array_equal(transformer.transform(df), transformed_arr)
        np.testing.assert_array_equal(transformer.transform(arr), transformed_arr)
        np.testing.assert_array_equal(transformer.transform(new_df), new_transformed_arr)
        pd.testing.assert_frame_equal(df, df_copy)
        pd.testing.assert_frame_equal(new_df, new_df_copy)
        np.testing.assert_array_equal(arr, arr_copy)

        # Strategy threshold
        transformed_arr_strat_threshold = pd.DataFrame({
            'x': ['a', 'b', 'b', 'c'] * 1000 + ['other_', 'other_', 'other_'] * 10,
            'y': ['a', 'b', 'b', 'c'] * 1006 + ['other_', 'other_', 'other_'] * 2,
            'z': ['a', 'b', 'b', 'c'] * 930 + ['other_'] * 200 + ['other_', 'other_'] * 55,
        }).to_numpy()
        transformer = column_preprocessors.AutoBinner(strategy="threshold", min_cat_count=3, threshold=0.05)
        transformer.fit(arr) # Fit on a numpy array to test while at it
        np.testing.assert_array_equal(transformer.transform(df), transformed_arr_strat_threshold)

        # min_cat_count > nb categories
        new_transformed_arr_min_cat_count_sup = pd.DataFrame({
            'toto': ['a', 'other_', 'b', 'c'] * 1000 + ['d', 'e', 'f'] * 10,
            'titi': ['d', 'd', 'e', 'f'] * 1006 + ['a', 'b', 'c'] * 2,
            'tata': ['other_', 'other_', 'b', 'c'] * 930 + ['d'] * 200 + ['e', 'f'] * 55,
        }).to_numpy()
        transformer = column_preprocessors.AutoBinner(strategy="auto", min_cat_count=100, threshold=0.05)
        transformer.fit(df)
        np.testing.assert_array_equal(transformer.transform(df), arr)
        np.testing.assert_array_equal(transformer.transform(new_df), new_transformed_arr_min_cat_count_sup)

        # Only one category under the threshold (keep it)
        df_2_cats = pd.DataFrame({'x': ['a']* 100 + ['b'] *2})
        arr_2_cats = pd.DataFrame({'x': ['a']* 100 + ['b'] *2}).to_numpy()
        transformer = column_preprocessors.AutoBinner(strategy="auto", min_cat_count=0, threshold=0.05)
        transformer.fit(df_2_cats)
        self.assertEqual(transformer.input_length, 1)
        np.testing.assert_array_equal(transformer.transform(df_2_cats), arr_2_cats)

        # Nominal case - fit_transform
        transformer = column_preprocessors.AutoBinner(strategy="auto", min_cat_count=3, threshold=0.05)
        self.assertFalse(hasattr(transformer, 'fitted_'))
        self.assertEqual(transformer.input_length, None)
        np.testing.assert_array_equal(transformer.fit_transform(df), transformed_arr)
        pd.testing.assert_frame_equal(df, df_copy)

        # Check errors
        with self.assertRaises(ValueError):
            column_preprocessors.AutoBinner(strategy='toto')
        with self.assertRaises(ValueError):
            column_preprocessors.AutoBinner(min_cat_count=-1)
        with self.assertRaises(ValueError):
            column_preprocessors.AutoBinner(threshold=-1)
        transformer = column_preprocessors.AutoBinner(strategy="threshold", min_cat_count=3, threshold=0.05)
        transformer.fit(df)

    def test04_EmbeddingTransformer(self):
        '''Test of the class column_preprocessors.EmbeddingTransformer'''
        # Values to test
        embedding = {'toto': [1., 5., 8.], 'titi': [-1., 3., 4.], 'tata': [-8., -6., 2.64]}
        df = pd.DataFrame({
            'x': ['toto', 'tata', 'titi'],
            'y': ['titi', 'test', 'toto'],
        })
        df_copy = df.copy(deep=True)
        arr = df.to_numpy()
        arr_copy = arr.copy()
        transformed_arr = pd.DataFrame({
            'x1': [1., -8., -1.],
            'x2': [5., -6., 3.],
            'x3': [8., 2.64, 4.],
            'y1': [-1., 0., 1.],
            'y2': [3., 0., 5.],
            'y3': [4., 0., 8.],
        }).to_numpy()

        new_df =  pd.DataFrame({
            'x': ['test', 'tata', 'tata'],
            'y': ['titi', 'tata', 'toto'],
        })
        new_df_copy = new_df.copy(deep=True)
        new_transformed_arr = pd.DataFrame({
            'x1': [0., -8., -8.],
            'x2': [0., -6., -6.],
            'x3': [0., 2.64, 2.64],
            'y1': [-1., -8., 1.],
            'y2': [3., -6., 5.],
            'y3': [4., 2.64, 8.],
        }).to_numpy()


        # Nominal case
        transformer = column_preprocessors.EmbeddingTransformer(embedding=embedding)
        transformer.fit(df)
        self.assertTrue(hasattr(transformer, 'fitted_'))
        self.assertEqual(transformer.input_length, 2)
        self.assertEqual(transformer.embedding_size, 3)
        np.testing.assert_array_equal(transformer.transform(df), transformed_arr)
        np.testing.assert_array_equal(transformer.transform(arr), transformed_arr)
        np.testing.assert_array_equal(transformer.transform(new_df), new_transformed_arr)
        pd.testing.assert_frame_equal(df, df_copy)
        pd.testing.assert_frame_equal(new_df, new_df_copy)
        np.testing.assert_array_equal(arr, arr_copy)

        # Nominal case - fit_transform
        transformer = column_preprocessors.EmbeddingTransformer(embedding=embedding)
        self.assertEqual(transformer.input_length, None)
        self.assertEqual(transformer.embedding_size, 3)
        np.testing.assert_array_equal(transformer.fit_transform(df), transformed_arr)
        pd.testing.assert_frame_equal(df, df_copy)

        # JSON file
        json_path = os.path.join(os.getcwd(), 'tmp_json_tests.json')
        if os.path.exists(json_path):
            os.remove(json_path)
        with open(json_path, 'w', encoding='{{default_encoding}}') as f:
            json.dump(embedding, f, indent=4)
        transformer = column_preprocessors.EmbeddingTransformer(embedding=json_path)
        transformer.fit(df)
        self.assertTrue(hasattr(transformer, 'fitted_'))
        self.assertEqual(transformer.input_length, 2)
        self.assertEqual(transformer.embedding_size, 3)
        np.testing.assert_array_equal(transformer.transform(df), transformed_arr)
        np.testing.assert_array_equal(transformer.transform(arr), transformed_arr)
        np.testing.assert_array_equal(transformer.transform(new_df), new_transformed_arr)
        pd.testing.assert_frame_equal(df, df_copy)
        pd.testing.assert_frame_equal(new_df, new_df_copy)
        np.testing.assert_array_equal(arr, arr_copy)
        os.remove(json_path)

        # get_feature_names
        transformer = column_preprocessors.EmbeddingTransformer(embedding=embedding)
        transformer.fit(arr) # Fit on a numpy array to test while at it
        features_out = transformer.get_feature_names(['toto', 'test'])
        expected_features_out = np.array([
            'emb_toto_0', 'emb_toto_1', 'emb_toto_2',
            'emb_test_0', 'emb_test_1', 'emb_test_2',
        ])
        np.testing.assert_array_equal(features_out, expected_features_out)

        # Check errors
        with self.assertRaises(ValueError):
            column_preprocessors.EmbeddingTransformer(embedding=embedding, none_strategy='toto')
        with self.assertRaises(ValueError):
            column_preprocessors.EmbeddingTransformer(embedding='toto.csv')
        with self.assertRaises(FileNotFoundError):
            column_preprocessors.EmbeddingTransformer(embedding='not_a_correct_path.json')
        transformer = column_preprocessors.EmbeddingTransformer(embedding=embedding)
        transformer.fit(df)


# Execute tests
if __name__ == '__main__':
    # Start tests
    unittest.main()



#!/usr/bin/env python3
# Copyright (C) <2018-2022>  <Agence Data Services, DSI Pôle Emploi>
# 
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
# 
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

# Libs unittest
import unittest
from unittest.mock import Mock
from unittest.mock import patch

# Utils libs
import os
import json
import shutil
import numpy as np
import pandas as pd

from sklearn.pipeline import Pipeline
from sklearn.ensemble import RandomForestClassifier, RandomForestRegressor

from {{package_name}} import utils
from {{package_name}}.models_training.model_pipeline import ModelPipeline

# Disable logging
import logging
logging.disable(logging.CRITICAL)


def remove_dir(path):
    if os.path.isdir(path): shutil.rmtree(path)


class ModelPipelineTests(unittest.TestCase):
    '''Main class to test model_pipeline'''


    def setUp(self):
        '''SetUp fonction'''
        # Change directory to script directory
        abspath = os.path.abspath(__file__)
        dname = os.path.dirname(abspath)
        os.chdir(dname)


    def test01_model_pipeline_init(self):
        '''Test of the initialization of {{package_name}}.models_training.model_pipeline.ModelPipeline'''

        model_dir = os.path.join(os.getcwd(), 'model_test_123456789')
        remove_dir(model_dir)

        # Init., test all parameters
        model = ModelPipeline(model_dir=model_dir)
        self.assertEqual(model.model_dir, model_dir)
        self.assertEqual(model.model_type, None)
        self.assertTrue(os.path.isdir(model_dir))
        self.assertEqual(model.pipeline, None)
        # We test display_if_gpu_activated and _is_gpu_activated just by calling them
        self.assertTrue(type(model._is_gpu_activated()) == bool)
        model.display_if_gpu_activated()
        remove_dir(model_dir)

        random_forest = RandomForestClassifier()
        pipeline = Pipeline([('rf', random_forest)])
        model = ModelPipeline(model_dir=model_dir, pipeline=pipeline)
        model.save()
        self.assertEqual(model.pipeline, pipeline)
        remove_dir(model_dir)


    def test02_model_pipeline_fit(self):
        '''Test of the method fit of {{package_name}}.models_training.model_pipeline.ModelPipeline'''

        model_dir = os.path.join(os.getcwd(), 'model_test_123456789')
        remove_dir(model_dir)

        # Set vars
        x_train = pd.DataFrame({'col_1': [-5, -1, 0, -2, 2, -6, 3] * 10, 'col_2': [2, -1, -8, 2, 3, 12, 2] * 10})
        y_train_mono_2 = pd.Series([0, 0, 0, 0, 1, 1, 1] * 10)
        y_train_mono_3 = pd.Series([0, 0, 0, 2, 1, 1, 1] * 10)
        y_train_regressor = pd.Series([-3, -2, -8, 0, 5, 6, 5] * 10)
        y_train_multi = pd.DataFrame({'y1': [0, 0, 0, 0, 1, 1, 1] * 10, 'y2': [1, 0, 0, 1, 1, 1, 1] * 10, 'y3': [0, 0, 1, 0, 1, 0, 1] * 10})
        x_col = ['col_1', 'col_2']
        y_col_mono = ['toto']
        y_col_multi = ['y1', 'y2', 'y3']

        # Classification - Mono-label - Mono-Class
        rf_classifier = RandomForestClassifier()
        pipeline = Pipeline([('rf', rf_classifier)])
        model = ModelPipeline(x_col=x_col, y_col=y_col_mono, model_dir=model_dir, pipeline=pipeline)
        # We mock a mono-label classifier
        model.model_type = 'classifier'
        model.multi_label = False
        self.assertFalse(model.trained)
        self.assertEqual(model.nb_fit, 0)
        self.assertEqual(model.x_col, x_col)
        self.assertEqual(model.y_col, y_col_mono)
        self.assertFalse(hasattr(model.pipeline['rf'], "classes_"))
        model.fit(x_train, y_train_mono_2)
        self.assertTrue(model.trained)
        self.assertEqual(model.nb_fit, 1)
        self.assertEqual(model.x_col, x_col)
        self.assertEqual(model.y_col, y_col_mono)
        self.assertTrue(hasattr(model.pipeline['rf'], "classes_"))
        self.assertEqual(model.list_classes, [0, 1])
        self.assertEqual(model.dict_classes, {0: 0, 1: 1})
        remove_dir(model_dir)

        # Classification - Mono-label - Multi-Classes
        rf_classifier = RandomForestClassifier()
        pipeline = Pipeline([('rf', rf_classifier)])
        model = ModelPipeline(x_col=x_col, y_col=y_col_mono, model_dir=model_dir, pipeline=pipeline)
        # We mock a mono-label classifier
        model.model_type = 'classifier'
        model.multi_label = False
        self.assertFalse(model.trained)
        self.assertEqual(model.nb_fit, 0)
        self.assertEqual(model.x_col, x_col)
        self.assertEqual(model.y_col, y_col_mono)
        self.assertFalse(hasattr(model.pipeline['rf'], "classes_"))
        model.fit(x_train, y_train_mono_3)
        self.assertTrue(model.trained)
        self.assertEqual(model.nb_fit, 1)
        self.assertEqual(model.x_col, x_col)
        self.assertEqual(model.y_col, y_col_mono)
        self.assertTrue(hasattr(model.pipeline['rf'], "classes_"))
        self.assertEqual(model.list_classes, [0, 1, 2])
        self.assertEqual(model.dict_classes, {0: 0, 1: 1, 2: 2})
        remove_dir(model_dir)

        # Classification - Multi-labels
        rf_classifier = RandomForestClassifier()
        pipeline = Pipeline([('rf', rf_classifier)])
        # We also check without x_col & y_col
        model = ModelPipeline(model_dir=model_dir, pipeline=pipeline)
        # We mock a multi-labels classifier
        model.model_type = 'classifier'
        model.multi_label = True
        self.assertFalse(model.trained)
        self.assertEqual(model.nb_fit, 0)
        self.assertEqual(model.x_col, None)
        self.assertEqual(model.y_col, None)
        self.assertFalse(hasattr(model.pipeline, "classes_"))
        model.fit(x_train, y_train_multi)
        self.assertTrue(model.trained)
        self.assertEqual(model.nb_fit, 1)
        self.assertEqual(model.x_col, x_col)
        self.assertEqual(model.y_col, y_col_multi)
        self.assertTrue(hasattr(model.pipeline, "classes_"))
        self.assertEqual(model.list_classes, y_col_multi)
        self.assertEqual(model.dict_classes, {0: 'y1', 1: 'y2', 2: 'y3'})
        remove_dir(model_dir)

        # Regressor
        rf_regressor = RandomForestRegressor()
        pipeline = Pipeline([('rf', rf_regressor)])
        model = ModelPipeline(x_col=x_col, y_col=y_col_mono, model_dir=model_dir, pipeline=pipeline)
        # We mock a regressor model
        model.model_type = 'regressor'
        self.assertFalse(model.trained)
        self.assertEqual(model.nb_fit, 0)
        self.assertEqual(model.x_col, x_col)
        self.assertEqual(model.y_col, y_col_mono)
        self.assertFalse(hasattr(model.pipeline['rf'], "feature_importances_"))
        model.fit(x_train, y_train_regressor)
        self.assertTrue(model.trained)
        self.assertEqual(model.nb_fit, 1)
        self.assertEqual(model.x_col, x_col)
        self.assertEqual(model.y_col, y_col_mono)
        self.assertTrue(hasattr(model.pipeline['rf'], "feature_importances_"))
        self.assertFalse(hasattr(model, "list_classes"))
        self.assertFalse(hasattr(model, "dict_classes"))
        remove_dir(model_dir)
        #
        ############
        # Test continue training
        rf_classifier = RandomForestClassifier()
        pipeline = Pipeline([('rf', rf_classifier)])
        model = ModelPipeline(x_col=x_col, y_col=y_col_mono, model_dir=model_dir, pipeline=pipeline)
        # We mock a mono-label classifier
        model.model_type = 'classifier'
        model.multi_label = False
        self.assertFalse(model.trained)
        self.assertEqual(model.nb_fit, 0)
        self.assertFalse(hasattr(model.pipeline['rf'], "classes_"))
        model.fit(x_train, y_train_mono_2)
        self.assertTrue(model.trained)
        self.assertEqual(model.nb_fit, 1)
        self.assertTrue(hasattr(model.pipeline['rf'], "classes_"))
        # Second fit
        with self.assertRaises(RuntimeError):
            model.fit(x_train[:50], y_train_mono_2[:50])
        self.assertEqual(model_dir, model.model_dir)
        remove_dir(model_dir)

        # Check errors
        rf_classifier = RandomForestClassifier()
        pipeline = Pipeline([('rf', rf_classifier)])
        model = ModelPipeline(x_col=x_col, y_col=y_col_mono, model_dir=model_dir, pipeline=pipeline)
        with self.assertRaises(ValueError):
            model.fit(x_train, y_train_mono_2)
        remove_dir(model_dir)


    def test03_model_pipeline_predict(self):
        '''Test of the method predict of {{package_name}}.models_training.model_pipeline.ModelPipeline'''

        model_dir = os.path.join(os.getcwd(), 'model_test_123456789')
        remove_dir(model_dir)

        # Set vars
        x_train = pd.DataFrame({'col_1': [-5, -1, 0, -2, 2, -6, 3] * 10, 'col_2': [2, -1, -8, 2, 3, 12, 2] * 10})
        y_train_mono_2 = pd.Series([0, 0, 0, 0, 1, 1, 1] * 10)
        y_train_mono_3 = pd.Series([0, 0, 0, 2, 1, 1, 1] * 10)
        y_train_regressor = pd.Series([-3, -2, -8, 0, 5, 6, 5] * 10)
        y_train_multi = pd.DataFrame({'y1': [0, 0, 0, 0, 1, 1, 1] * 10, 'y2': [1, 0, 0, 1, 1, 1, 1] * 10, 'y3': [0, 0, 1, 0, 1, 0, 1] * 10})
        x_col = ['col_1', 'col_2']
        y_col_mono = ['toto']
        y_col_multi = ['y1', 'y2', 'y3']

        # Classification - Mono-label - Mono-Class
        rf_classifier = RandomForestClassifier(n_estimators=10)
        pipeline = Pipeline([('rf', rf_classifier)])
        model = ModelPipeline(x_col=x_col, y_col=y_col_mono, model_dir=model_dir, pipeline=pipeline)
        # We mock a mono-label classifier
        model.model_type = 'classifier'
        model.multi_label = False
        model.fit(x_train, y_train_mono_2)
        preds = model.predict(x_train, return_proba=False)
        self.assertEqual(preds.shape, (len(x_train),))
        proba = model.predict(x_train, return_proba=True)
        self.assertEqual(proba.shape, (len(x_train), 2)) # 2 classes
        remove_dir(model_dir)

        # Classification - Mono-label - Multi-Classes
        rf_classifier = RandomForestClassifier(n_estimators=10)
        pipeline = Pipeline([('rf', rf_classifier)])
        model = ModelPipeline(x_col=x_col, y_col=y_col_mono, model_dir=model_dir, pipeline=pipeline)
        # We mock a mono-label classifier
        model.model_type = 'classifier'
        model.multi_label = False
        model.fit(x_train, y_train_mono_3)
        preds = model.predict(x_train, return_proba=False)
        self.assertEqual(preds.shape, (len(x_train),))
        proba = model.predict(x_train, return_proba=True)
        self.assertEqual(proba.shape, (len(x_train), 3)) # 3 classes
        remove_dir(model_dir)

        # Classification - Multi-labels
        rf_classifier = RandomForestClassifier(n_estimators=10)
        pipeline = Pipeline([('rf', rf_classifier)])
        model = ModelPipeline(x_col=x_col, y_col=y_col_multi, model_dir=model_dir, pipeline=pipeline)
        # We mock a multi-labels classifier
        model.model_type = 'classifier'
        model.multi_label = True
        model.fit(x_train, y_train_multi)
        preds = model.predict(x_train)
        self.assertEqual(preds.shape, (len(x_train), len(y_col_multi)))
        proba = model.predict(x_train, return_proba=True)
        self.assertEqual(proba.shape, (len(x_train), len(y_col_multi)))
        remove_dir(model_dir)

        # Regressor
        rf_regressor = RandomForestRegressor(n_estimators=10)
        pipeline = Pipeline([('rf', rf_regressor)])
        model = ModelPipeline(x_col=x_col, y_col=y_col_mono, model_dir=model_dir, pipeline=pipeline)
        # We mock a regressor model
        model.model_type = 'regressor'
        model.fit(x_train, y_train_regressor)
        preds = model.predict(x_train)
        self.assertEqual(preds.shape, (len(x_train),))
        with self.assertRaises(ValueError):
            proba = model.predict(x_train, return_proba=True)
        remove_dir(model_dir)

        # Model needs to be fitted
        with self.assertRaises(AttributeError):
            model = ModelPipeline(x_col=x_col, y_col=y_col_mono, model_dir=model_dir, pipeline=pipeline)
            model.predict(pd.Series([-2, 3]))
        remove_dir(model_dir)


    def test04_model_pipeline_predict_proba(self):
        '''Test of the method predict_proba of {{package_name}}.models_training.model_pipeline.ModelPipeline'''

        model_dir = os.path.join(os.getcwd(), 'model_test_123456789')
        remove_dir(model_dir)

        # Set vars
        x_train = pd.DataFrame({'col_1': [-5, -1, 0, -2, 2, -6, 3] * 10, 'col_2': [2, -1, -8, 2, 3, 12, 2] * 10})
        y_train_mono_2 = pd.Series([0, 0, 0, 0, 1, 1, 1] * 10)
        y_train_mono_3 = pd.Series([0, 0, 0, 2, 1, 1, 1] * 10)
        y_train_regressor = pd.Series([-3, -2, -8, 0, 5, 6, 5] * 10)
        y_train_multi = pd.DataFrame({'y1': [0, 0, 0, 0, 1, 1, 1] * 10, 'y2': [1, 0, 0, 1, 1, 1, 1] * 10, 'y3': [0, 0, 1, 0, 1, 0, 1] * 10})
        x_col = ['col_1', 'col_2']
        y_col_mono = ['toto']
        y_col_multi = ['y1', 'y2', 'y3']

        # Classification - Mono-label - Mono-Class
        rf_classifier = RandomForestClassifier(n_estimators=10)
        pipeline = Pipeline([('rf', rf_classifier)])
        model = ModelPipeline(x_col=x_col, y_col=y_col_mono, model_dir=model_dir, pipeline=pipeline)
        # We mock a mono-label classifier
        model.model_type = 'classifier'
        model.multi_label = False
        model.fit(x_train, y_train_mono_2)
        preds = model.predict_proba(x_train)
        self.assertEqual(preds.shape, (len(x_train), 2)) # 2 classes
        self.assertTrue(isinstance(preds[0][0], (np.floating, float)))
        remove_dir(model_dir)

        # Classification - Mono-label - Multi-Classes
        rf_classifier = RandomForestClassifier(n_estimators=10)
        pipeline = Pipeline([('rf', rf_classifier)])
        model = ModelPipeline(x_col=x_col, y_col=y_col_mono, model_dir=model_dir, pipeline=pipeline)
        # We mock a mono-label classifier
        model.model_type = 'classifier'
        model.multi_label = False
        model.fit(x_train, y_train_mono_3)
        preds = model.predict_proba(x_train)
        self.assertEqual(preds.shape, (len(x_train), 3)) # 3 classes
        self.assertTrue(isinstance(preds[0][0], (np.floating, float)))
        remove_dir(model_dir)

        # Classification - Multi-labels
        rf_classifier = RandomForestClassifier(n_estimators=10)
        pipeline = Pipeline([('rf', rf_classifier)])
        model = ModelPipeline(x_col=x_col, y_col=y_col_multi, model_dir=model_dir, pipeline=pipeline)
        # We mock a multi-labels classifier
        model.model_type = 'classifier'
        model.multi_label = True
        model.fit(x_train, y_train_multi)
        preds = model.predict_proba(x_train)
        self.assertEqual(preds.shape, (len(x_train), len(y_col_multi))) # 3 labels
        self.assertTrue(isinstance(preds[0][0], (np.floating, float)))
        remove_dir(model_dir)

        # Regressor
        rf_regressor = RandomForestRegressor(n_estimators=10)
        pipeline = Pipeline([('rf', rf_regressor)])
        model = ModelPipeline(x_col=x_col, y_col=y_col_mono, model_dir=model_dir, pipeline=pipeline)
        # We mock a regressor model
        model.model_type = 'regressor'
        model.fit(x_train, y_train_regressor)
        with self.assertRaises(ValueError):
            proba = model.predict_proba(x_train)
        remove_dir(model_dir)

        # Model needs to be fitted
        with self.assertRaises(AttributeError):
            model = ModelPipeline(x_col=x_col, y_col=y_col_mono, model_dir=model_dir, pipeline=pipeline)
            model.predict_proba('test')
        remove_dir(model_dir)


    def test05_model_pipeline_save(self):
        '''Test of the method save of {{package_name}}.models_training.model_pipeline.ModelPipeline'''

        model_dir = os.path.join(os.getcwd(), 'model_test_123456789')
        remove_dir(model_dir)
        x_col = ['col_1', 'col_2']
        y_col_mono = ['toto']

        # With Pipeline - Classifier
        rf_classifier = RandomForestClassifier(n_estimators=10)
        pipeline = Pipeline([('rf', rf_classifier)])
        model = ModelPipeline(x_col=x_col, y_col=y_col_mono, model_dir=model_dir, pipeline=pipeline)
        model.model_type = 'classifier'
        # Save
        model.save(json_data={'test': 8})
        # Assert everything has been saved
        self.assertTrue(os.path.exists(os.path.join(model.model_dir, 'configurations.json')))
        self.assertTrue(os.path.exists(os.path.join(model.model_dir, f"{model.model_name}.pkl")))
        self.assertTrue(os.path.exists(os.path.join(model.model_dir, f"sklearn_pipeline_standalone.pkl")))
        self.assertTrue(os.path.exists(os.path.join(model.model_dir, 'preprocess_pipeline.pkl')))
        with open(os.path.join(model.model_dir, 'configurations.json'), 'r', encoding='{{default_encoding}}') as f:
            configs = json.load(f)
        self.assertEqual(configs['test'], 8)
        self.assertTrue('mainteners' in configs.keys())
        self.assertTrue('date' in configs.keys())
        self.assertTrue('package_version' in configs.keys())
        self.assertEqual(configs['package_version'], utils.get_package_version())
        self.assertTrue('model_name' in configs.keys())
        self.assertTrue('model_dir' in configs.keys())
        self.assertTrue('model_type' in configs.keys())
        self.assertEqual(configs['model_type'], 'classifier')
        self.assertTrue('trained' in configs.keys())
        self.assertTrue('nb_fit' in configs.keys())
        self.assertTrue('x_col' in configs.keys())
        self.assertTrue('y_col' in configs.keys())
        self.assertTrue('columns_in' in configs.keys())
        self.assertTrue('mandatory_columns' in configs.keys())
        self.assertTrue('level_save' in configs.keys())
        self.assertTrue('librairie' in configs.keys())
        self.assertEqual(configs['librairie'], 'scikit-learn')
        self.assertTrue('list_classes' not in configs.keys()) # not in because we do not use the Classifier mixin
        self.assertTrue('dict_classes' not in configs.keys()) # not in because we do not use the Classifier mixin
        self.assertTrue('multi_label' not in configs.keys()) # not in because we do not use the Classifier mixin
        # Specific model used
        self.assertTrue('rf_confs' in configs.keys())
        remove_dir(model_dir)

        # With Pipeline - Regressor
        rf_regressor = RandomForestRegressor(n_estimators=10)
        pipeline = Pipeline([('rf', rf_regressor)])
        model = ModelPipeline(x_col=x_col, y_col=y_col_mono, model_dir=model_dir, pipeline=pipeline)
        model.model_type = 'regressor'
        # Save
        model.save(json_data={'test': 8})
        # Assert everything has been saved
        self.assertTrue(os.path.exists(os.path.join(model.model_dir, 'configurations.json')))
        self.assertTrue(os.path.exists(os.path.join(model.model_dir, f"{model.model_name}.pkl")))
        self.assertTrue(os.path.exists(os.path.join(model.model_dir, f"sklearn_pipeline_standalone.pkl")))
        self.assertTrue(os.path.exists(os.path.join(model.model_dir, 'preprocess_pipeline.pkl')))
        with open(os.path.join(model.model_dir, 'configurations.json'), 'r', encoding='{{default_encoding}}') as f:
            configs = json.load(f)
        self.assertEqual(configs['test'], 8)
        self.assertTrue('mainteners' in configs.keys())
        self.assertTrue('date' in configs.keys())
        self.assertTrue('package_version' in configs.keys())
        self.assertEqual(configs['package_version'], utils.get_package_version())
        self.assertTrue('model_name' in configs.keys())
        self.assertTrue('model_dir' in configs.keys())
        self.assertTrue('model_type' in configs.keys())
        self.assertEqual(configs['model_type'], 'regressor')
        self.assertTrue('trained' in configs.keys())
        self.assertTrue('nb_fit' in configs.keys())
        self.assertTrue('x_col' in configs.keys())
        self.assertTrue('y_col' in configs.keys())
        self.assertTrue('columns_in' in configs.keys())
        self.assertTrue('mandatory_columns' in configs.keys())
        self.assertTrue('level_save' in configs.keys())
        self.assertTrue('librairie' in configs.keys())
        self.assertEqual(configs['librairie'], 'scikit-learn')
        self.assertTrue('list_classes' not in configs.keys()) # Not in because Regressor
        self.assertTrue('dict_classes' not in configs.keys()) # Not in because Regressor
        self.assertTrue('multi_label' not in configs.keys()) # Not in because Regressor
        # Specific model used
        self.assertTrue('rf_confs' in configs.keys())
        remove_dir(model_dir)

        # Without Pipeline
        model = ModelPipeline(x_col=x_col, y_col=y_col_mono, model_dir=model_dir, pipeline=None)
        model.model_type = 'classifier' # We do not test for regressor, it is the same thing
        model.save(json_data={'test': 8})
        self.assertTrue(os.path.exists(os.path.join(model.model_dir, 'configurations.json')))
        self.assertTrue(os.path.exists(os.path.join(model.model_dir, f"{model.model_name}.pkl")))
        self.assertFalse(os.path.exists(os.path.join(model.model_dir, f"sklearn_pipeline_standalone.pkl")))
        self.assertTrue(os.path.exists(os.path.join(model.model_dir, 'preprocess_pipeline.pkl')))
        with open(os.path.join(model.model_dir, 'configurations.json'), 'r', encoding='{{default_encoding}}') as f:
            configs = json.load(f)
        self.assertEqual(configs['test'], 8)
        self.assertTrue('mainteners' in configs.keys())
        self.assertTrue('date' in configs.keys())
        self.assertTrue('package_version' in configs.keys())
        self.assertEqual(configs['package_version'], utils.get_package_version())
        self.assertTrue('model_name' in configs.keys())
        self.assertTrue('model_dir' in configs.keys())
        self.assertTrue('model_type' in configs.keys())
        self.assertEqual(configs['model_type'], 'classifier')
        self.assertTrue('trained' in configs.keys())
        self.assertTrue('nb_fit' in configs.keys())
        self.assertTrue('x_col' in configs.keys())
        self.assertTrue('y_col' in configs.keys())
        self.assertTrue('columns_in' in configs.keys())
        self.assertTrue('mandatory_columns' in configs.keys())
        self.assertTrue('level_save' in configs.keys())
        self.assertTrue('librairie' in configs.keys())
        self.assertEqual(configs['librairie'], 'scikit-learn')
        self.assertTrue('list_classes' not in configs.keys()) # not in because we do not use the Classifier mixin
        self.assertTrue('dict_classes' not in configs.keys()) # not in because we do not use the Classifier mixin
        self.assertTrue('multi_label' not in configs.keys()) # not in because we do not use the Classifier mixin
        # Specific model used
        self.assertTrue('rf_confs' not in configs.keys())
        remove_dir(model_dir)

        # WITH level_save = 'LOW'
        rf_classifier = RandomForestClassifier(n_estimators=10)
        pipeline = Pipeline([('rf', rf_classifier)])
        model = ModelPipeline(x_col=x_col, y_col=y_col_mono, model_dir=model_dir, pipeline=pipeline, level_save='LOW')
        model.model_type = 'classifier'
        model.multi_label = False
        # Save
        model.save(json_data={'test': 8})
        self.assertTrue(os.path.exists(os.path.join(model.model_dir, 'configurations.json')))
        self.assertFalse(os.path.exists(os.path.join(model.model_dir, f"{model.model_name}.pkl")))
        self.assertFalse(os.path.exists(os.path.join(model.model_dir, f"sklearn_pipeline_standalone.pkl")))
        self.assertFalse(os.path.exists(os.path.join(model.model_dir, 'preprocess_pipeline.pkl')))
        with open(os.path.join(model.model_dir, 'configurations.json'), 'r', encoding='{{default_encoding}}') as f:
            configs = json.load(f)
        self.assertEqual(configs['test'], 8)
        self.assertTrue('mainteners' in configs.keys())
        self.assertTrue('date' in configs.keys())
        self.assertTrue('package_version' in configs.keys())
        self.assertEqual(configs['package_version'], utils.get_package_version())
        self.assertTrue('model_name' in configs.keys())
        self.assertTrue('model_dir' in configs.keys())
        self.assertTrue('model_type' in configs.keys())
        self.assertEqual(configs['model_type'], 'classifier')
        self.assertTrue('trained' in configs.keys())
        self.assertTrue('nb_fit' in configs.keys())
        self.assertTrue('x_col' in configs.keys())
        self.assertTrue('y_col' in configs.keys())
        self.assertTrue('columns_in' in configs.keys())
        self.assertTrue('mandatory_columns' in configs.keys())
        self.assertTrue('level_save' in configs.keys())
        self.assertTrue('librairie' in configs.keys())
        self.assertEqual(configs['librairie'], 'scikit-learn')
        self.assertTrue('list_classes' not in configs.keys()) # not in because we do not use the Classifier mixin
        self.assertTrue('dict_classes' not in configs.keys()) # not in because we do not use the Classifier mixin
        self.assertTrue('multi_label' not in configs.keys()) # not in because we do not use the Classifier mixin
        # Specific model used
        self.assertTrue('rf_confs' in configs.keys())
        remove_dir(model_dir)


# Perform tests
if __name__ == '__main__':
    # Start tests
    unittest.main()



#!/usr/bin/env python3
# Copyright (C) <2018-2022>  <Agence Data Services, DSI Pôle Emploi>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

# Libs unittest
import unittest
from unittest.mock import Mock
from unittest.mock import patch

# Utils libs
import os
import json
import shutil
import numpy as np
import pandas as pd

from {{package_name}} import utils
from {{package_name}}.preprocessing import preprocess
from {{package_name}}.monitoring.model_logger import ModelLogger
from {{package_name}}.models_training.model_class import ModelClass
from {{package_name}}.models_training.classifiers.model_classifier import ModelClassifierMixin

# Disable logging
import logging
logging.disable(logging.CRITICAL)


def remove_dir(path):
    if os.path.isdir(path): shutil.rmtree(path)


class ModelMockClassifier(ModelClassifierMixin, ModelClass):
    '''We need a mock implementation of the Mixin class'''
    def __init__(self, **kwargs) -> None:
        super().__init__(**kwargs)
    def fit(self, df_train: pd.DataFrame, df_valid: pd.DataFrame = None, with_shuffle: bool = True, **kwargs):
        '''Simplified version of fit'''
        self.list_classes = sorted(list(df_train['file_class'].unique()))
        self.dict_classes = {i: col for i, col in enumerate(self.list_classes)}
        self.trained = True
        self.nb_fit += 1
    def predict(self, df_test, return_proba: bool = False, batch_size: int = None):
        '''Simplified version of predict'''
        predicted_proba = []
        for i, row in df_test.iterrows():
            tmp_probas = self.predict_on_name(row['file_path'])
            predicted_proba.append(tmp_probas)
        predicted_proba = np.array(predicted_proba)
        if return_proba:
            return predicted_proba
        else:
            return self.get_classes_from_proba(predicted_proba)
    def predict_proba(self, df_test, batch_size: int = None):
        '''Simplified version of predict_proba'''
        return self.predict(df_test, return_proba=True, batch_size=batch_size)
    def predict_on_name(self, name: str):
        if 'toto' in name:
            return [0.8, 0.1, 0.1]
        elif 'titi' in name:
            return [0.7, 0.1, 0.2]
        elif 'tata' in name:
            return [0.2, 0.6, 0.2]
        else:
            return [0.0, 0.2, 0.8]


class ModelClassifierMixinTests(unittest.TestCase):
    '''Main class to test model_classifier'''

    def setUp(self):
        '''SetUp fonction'''
        # Change directory to script directory
        abspath = os.path.abspath(__file__)
        dname = os.path.dirname(abspath)
        os.chdir(dname)

    def test01_model_classifier_init(self):
        '''Test of the initialization of {{package_name}}.models_training.classifiers.model_classifier.ModelClassifierMixin'''

        # Model creation
        model_dir = os.path.join(os.getcwd(), 'model_test_123456789')
        remove_dir(model_dir)
        model_name = 'test'

        # Nominal case
        model = ModelMockClassifier(model_dir=model_dir, model_name=model_name)
        self.assertEqual(model.model_dir, model_dir)
        self.assertTrue(os.path.isdir(model_dir))
        self.assertEqual(model.list_classes, None)
        self.assertEqual(model.dict_classes, None)
        self.assertEqual(model.model_type, 'classifier')
        remove_dir(model_dir)

        # Test level_save
        model = ModelMockClassifier(model_dir=model_dir, model_name=model_name, level_save='HIGH')
        self.assertEqual(model.level_save, 'HIGH')
        remove_dir(model_dir)
        #
        model = ModelMockClassifier(model_dir=model_dir, model_name=model_name, level_save='MEDIUM')
        self.assertEqual(model.level_save, 'MEDIUM')
        remove_dir(model_dir)
        #
        model = ModelMockClassifier(model_dir=model_dir, model_name=model_name, level_save='LOW')
        self.assertEqual(model.level_save, 'LOW')
        remove_dir(model_dir)

        # Manage errors
        with self.assertRaises(ValueError):
            ModelMockClassifier(model_dir=model_dir, model_name=model_name, level_save='toto')
        remove_dir(model_dir)

    def test02_model_classifier_predict_with_proba(self):
        '''Test of the method {{package_name}}.models_training.classifiers.model_classifier.ModelClassifierMixin.predict_with_proba'''

        # Model creation
        model_dir = os.path.join(os.getcwd(), 'model_test_123456789')
        remove_dir(model_dir)
        model_name = 'test'

        # Set vars
        df_train = pd.DataFrame({
            'file_class': ['cl1', 'cl1', 'cl2', 'cl3', 'cl1'],
            'file_path': ['toto.png', 'titi.png', 'tata.png', 'tutu.png', 'toto.png'],
        })

        # Nominal case
        model = ModelMockClassifier(model_dir=model_dir, model_name=model_name)
        model.fit(df_train)
        preds, probas = model.predict_with_proba(df_train)
        self.assertEqual(preds.shape, (df_train.shape[0],))
        self.assertEqual(probas.shape, (df_train.shape[0], 3))
        remove_dir(model_dir)

        # Model needs to be fitted
        with self.assertRaises(AttributeError):
            model = ModelMockClassifier(model_dir=model_dir, model_name=model_name)
            model.predict_with_proba(df_train)
        remove_dir(model_dir)

    def test03_model_classifier_get_predict_position(self):
        '''Test of the method {{package_name}}.models_training.classifiers.model_classifier.ModelClassifierMixin.get_predict_position'''

        # Model creation
        model_dir = os.path.join(os.getcwd(), 'model_test_123456789')
        remove_dir(model_dir)
        model_name = 'test'

        # Set vars
        df_train = pd.DataFrame({
            'file_class': ['cl1', 'cl1', 'cl2', 'cl3', 'cl1'],
            'file_path': ['toto.png', 'titi.png', 'tata.png', 'tutu.png', 'toto.png'],
        })
        y_true = ['cl1', 'cl1', 'cl1', 'cl3', 'cl2']  # Add some errors

        # Nominal case
        model = ModelMockClassifier(model_dir=model_dir, model_name=model_name)
        model.fit(df_train)
        predict_positions = model.get_predict_position(df_train, y_true)
        self.assertEqual(predict_positions.shape, (df_train.shape[0],))
        remove_dir(model_dir)

        # Model needs to be fitted
        with self.assertRaises(AttributeError):
            model = ModelMockClassifier(model_dir=model_dir, model_name=model_name)
            model.get_predict_position(df_train, y_true)
        remove_dir(model_dir)

    def test04_model_classifier_get_classes_from_proba(self):
        '''Test of the method {{package_name}}.models_training.classifiers.model_classifier.ModelClassifierMixin.get_classes_from_proba'''

        # Model creation
        model_dir = os.path.join(os.getcwd(), 'model_test_123456789')
        remove_dir(model_dir)
        model_name = 'test'

        # Set vars
        df_train = pd.DataFrame({
            'file_class': ['cl1', 'cl1', 'cl2', 'cl3', 'cl1'],
            'file_path': ['toto.png', 'titi.png', 'tata.png', 'tutu.png', 'toto.png'],
        })

        # Nominal case
        model = ModelMockClassifier(model_dir=model_dir, model_name=model_name)
        model.fit(df_train)
        index_cl1 = model.list_classes.index('cl1')
        index_cl2 = model.list_classes.index('cl2')
        index_cl3 = model.list_classes.index('cl3')
        inv_cl = [0, 0, 0]
        inv_cl[index_cl1] = 'cl1'
        inv_cl[index_cl2] = 'cl2'
        inv_cl[index_cl3] = 'cl3'
        predicted_proba = np.array([[0.7, 0.2, 0.1], [0.1, 0.9, 0.0], [0.1, 0.2, 0.7]])
        predicted_classes = model.get_classes_from_proba(predicted_proba)
        self.assertEqual(predicted_classes.shape, (predicted_proba.shape[0], ))
        self.assertEqual(predicted_classes[0], inv_cl[0])
        self.assertEqual(predicted_classes[1], inv_cl[1])
        self.assertEqual(predicted_classes[2], inv_cl[2])
        remove_dir(model_dir)

    def test05_model_classifier_get_top_n_from_proba(self):
        '''Test of the method {{package_name}}.models_training.classifiers.model_classifier.ModelClassifierMixin.get_top_n_from_proba'''

        # Model creation
        model_dir = os.path.join(os.getcwd(), 'model_test_123456789')
        remove_dir(model_dir)
        model_name = 'test'

        # Nominal case
        model = ModelMockClassifier(model_dir=model_dir, model_name=model_name)
        model.list_classes = ['a', 'b', 'c']
        model.dict_classes = {0: 'a', 1: 'b', 2: 'c'}
        probas = np.array([[0.1, 0.6, 0.3], [0.7, 0.2, 0.1]])
        top_n, top_n_proba = model.get_top_n_from_proba(probas, n=2)
        self.assertEqual([list(_) for _ in top_n], [['b', 'c'], ['a', 'b']])
        self.assertEqual([list(_) for _ in top_n_proba], [[0.6, 0.3], [0.7, 0.2]])
        with self.assertRaises(ValueError):
            model.get_top_n_from_proba(probas, n=5)
        remove_dir(model_dir)

    def test06_model_classifier_inverse_transform(self):
        '''Test of the method {{package_name}}.models_training.classifiers.model_classifier.ModelClassifierMixin.inverse_transform'''

        # Model creation
        model_dir = os.path.join(os.getcwd(), 'model_test_123456789')
        remove_dir(model_dir)
        model_name = 'test'

        # Nominal case
        model = ModelMockClassifier(model_dir=model_dir, model_name=model_name)
        model.list_classes = ['toto', 'titi', 'tata']
        y1 = np.array(['toto', 'titi', 'tata', 'toto'])
        expected_result1 = ['toto', 'titi', 'tata', 'toto']
        y2 = 'toto'
        expected_result2 = 'toto'
        self.assertEqual(model.inverse_transform(y1), expected_result1)
        self.assertEqual(model.inverse_transform(y2), expected_result2)
        remove_dir(model_dir)

    def test07_model_classifier_get_and_save_metrics(self):
        '''Test of the method {{package_name}}.models_training.classifiers.model_classifier.ModelClassifierMixin.get_and_save_metrics'''

        # Model creation
        model_dir = os.path.join(os.getcwd(), 'model_test_123456789')
        remove_dir(model_dir)
        model_name = 'test'

        # Nominal case
        model = ModelMockClassifier(model_dir=model_dir, model_name=model_name)
        model.list_classes = [0, 1]
        y_true = np.array([0, 1, 0, 1])
        y_pred = np.array([0, 1, 1, 0])
        df_metrics = model.get_and_save_metrics(y_true, y_pred)
        self.assertEqual(df_metrics.shape[0], 3) # 2 classes + All
        self.assertEqual(df_metrics.loc[2, :]['Label'], 'All')
        self.assertEqual(df_metrics.loc[2, :]['Accuracy'], 0.5)
        plots_path = os.path.join(model.model_dir, 'plots')
        self.assertTrue(os.path.exists(os.path.join(model.model_dir, 'predictions.csv')))
        self.assertTrue(os.path.exists(os.path.join(plots_path, 'confusion_matrix_normalized.png')))
        self.assertTrue(os.path.exists(os.path.join(plots_path, 'confusion_matrix.png')))
        remove_dir(model_dir)

        # With the other parameters
        model = ModelMockClassifier(model_dir=model_dir, model_name=model_name)
        model.list_classes = [0, 1]
        y_true = np.array([0, 1, 0, 1])
        y_pred = np.array([0, 1, 1, 0])
        list_files_x = ['toto.png', 'titi.png', 'tata.png', 'tutu.png']
        type_data = 'toto'
        model_logger = ModelLogger(
            tracking_uri="http://toto.titi.tata.test",
            experiment_name="test"
        )
        df_metrics = model.get_and_save_metrics(y_true, y_pred, list_files_x=list_files_x, type_data=type_data, model_logger=model_logger)
        self.assertEqual(df_metrics.shape[0], 3) # 2 classes + All
        self.assertEqual(df_metrics.loc[2, :]['Label'], 'All')
        self.assertEqual(df_metrics.loc[2, :]['Accuracy'], 0.5)
        plots_path = os.path.join(model.model_dir, 'plots')
        self.assertTrue(os.path.exists(os.path.join(model.model_dir, f'predictions_{type_data}.csv')))
        self.assertTrue(os.path.exists(os.path.join(plots_path, f'{type_data}_confusion_matrix_normalized.png')))
        self.assertTrue(os.path.exists(os.path.join(plots_path, f'{type_data}_confusion_matrix.png')))
        df_preds = pd.read_csv(os.path.join(model.model_dir, f'predictions_{type_data}.csv'), sep='{{default_sep}}', encoding='{{default_encoding}}')
        self.assertTrue('y_true' in df_preds.columns)
        self.assertTrue('y_pred' in df_preds.columns)
        self.assertTrue('file_path' in df_preds.columns)
        self.assertTrue('matched' in df_preds.columns)
        remove_dir(model_dir)

    def test08_model_classifier_get_metrics_simple_monolabel(self):
        '''Test of the method {{package_name}}.models_training.classifiers.model_classifier.ModelClassifierMixin.get_metrics_simple_monolabel'''

        # Model creation
        model_dir = os.path.join(os.getcwd(), 'model_test_123456789')
        remove_dir(model_dir)
        model_name = 'test'

        # Nominal case
        model = ModelMockClassifier(model_dir=model_dir, model_name=model_name)
        model.list_classes = [0, 1]
        y_true = np.array([0, 1, 0, 1])
        y_pred = np.array([0, 1, 1, 0])
        df_metrics = model.get_metrics_simple_monolabel(y_true, y_pred)
        self.assertEqual(df_metrics.shape[0], 3) # 2 classes + All
        self.assertEqual(df_metrics.loc[2, :]['Label'], 'All')
        self.assertEqual(df_metrics.loc[2, :]['Accuracy'], 0.5)
        remove_dir(model_dir)

    def test09_model_classifier_update_info_from_c_mat(self):
        '''Test of the method {{package_name}}.models_training.classifiers.model_classifier.ModelClassifierMixin._update_info_from_c_mat'''

        # Model creation
        model_dir = os.path.join(os.getcwd(), 'model_test_123456789')
        remove_dir(model_dir)
        model_name = 'test'
        model = ModelMockClassifier(model_dir=model_dir, model_name=model_name)
        c_mat = [[10, 5], [8, 1]]
        expected_result = {'Accuracy': (10 + 1) / (10 + 5 + 8 + 1),
                           'Condition negative': 10 + 5,
                           'Condition positive': 8 + 1,
                           'F1-Score': 2 / ((1/(1 / (5 + 1))) + (1/(1 / (8 + 1)))),
                           'False negative': 8,
                           'False positive': 5,
                           'Falses': 5 + 8,
                           'Label': 'toto',
                           'Precision': 1 / (5 + 1),
                           'Predicted negative': 10 + 8,
                           'Predicted positive': 5 + 1,
                           'Recall': 1 / (8 + 1),
                           'True negative': 10,
                           'True positive': 1,
                           'Trues': 10 + 1}

        # Nominal case
        info_dict = model._update_info_from_c_mat(c_mat, label='toto', log_info=False)
        self.assertEqual(info_dict, expected_result)

    def test10_model_classifier_save(self):
        '''Test of the method {{package_name}}.models_training.classifiers.model_classifier.ModelClassifierMixin.save'''

        # Model creation
        model_dir = os.path.join(os.getcwd(), 'model_test_123456789')
        remove_dir(model_dir)
        model_name = 'test'

        # test save
        model = ModelMockClassifier(model_dir=model_dir, model_name=model_name)
        pkl_path = os.path.join(model.model_dir, f"{model.model_name}.pkl")
        configuration_path = os.path.join(model.model_dir, 'configurations.json')
        model.save(json_data={'test': 8})
        self.assertTrue(os.path.exists(pkl_path))
        self.assertTrue(os.path.exists(configuration_path))
        with open(configuration_path, 'r', encoding='{{default_encoding}}') as f:
            configs = json.load(f)
        self.assertEqual(configs['test'], 8)
        self.assertTrue('mainteners' in configs.keys())
        self.assertTrue('date' in configs.keys())
        self.assertTrue('package_version' in configs.keys())
        self.assertEqual(configs['package_version'], utils.get_package_version())
        self.assertTrue('model_name' in configs.keys())
        self.assertTrue('model_dir' in configs.keys())
        self.assertTrue('model_type' in configs.keys())
        self.assertTrue('trained' in configs.keys())
        self.assertTrue('nb_fit' in configs.keys())
        self.assertTrue('level_save' in configs.keys())
        self.assertTrue('librairie' in configs.keys())
        self.assertEqual(configs['librairie'], None)
        self.assertTrue('list_classes' in configs.keys())
        self.assertTrue('dict_classes' in configs.keys())
        remove_dir(model_dir)

        # test save, level_save = 'LOW'
        model = ModelMockClassifier(model_dir=model_dir, model_name=model_name, level_save='LOW')
        pkl_path = os.path.join(model.model_dir, f"{model.model_name}.pkl")
        configuration_path = os.path.join(model.model_dir, 'configurations.json')
        model.save(json_data={'test': 8})
        self.assertFalse(os.path.exists(pkl_path))
        self.assertTrue(os.path.exists(configuration_path))
        with open(configuration_path, 'r', encoding='{{default_encoding}}') as f:
            configs = json.load(f)
        self.assertEqual(configs['test'], 8)
        remove_dir(model_dir)


# Perform tests
if __name__ == '__main__':
    # Start tests
    unittest.main()



#!/usr/bin/env python3

## Preprocessing functions
# Copyright (C) <2018-2022>  <Agence Data Services, DSI Pôle Emploi>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


import os
import re
import json
import logging
import functools
import numpy as np
import pandas as pd
from sklearn.base import BaseEstimator
from typing import List, Dict, Tuple, Any, Union, Optional
from sklearn.utils.validation import check_is_fitted
from sklearn.preprocessing._function_transformer import FunctionTransformer

logger = logging.getLogger(__name__)


class Estimator(BaseEstimator):
    '''Base class for the classes defined below. Implements _validate_input and fit_transform.'''

    def __init__(self, input_length: Union[int, None]) -> None:
        '''Initialization of the class

        Args:
            input_length (int): The number of columns of the input (used in _validate_input())
        '''
        self.input_length = input_length

    def _validate_input(self, X: Union[np.ndarray, pd.DataFrame]) -> Union[np.ndarray, pd.DataFrame]:
        '''Validates input format

        Args:
            X (np.ndarray or pd.DataFrame): Input to validate
        Raises:
            ValueError: If the shape of the input does not correspond to self.input_length
        Returns:
            np.ndarray or pd.DataFrame: A copy of X
        '''
        if self.input_length is not None and X.shape[1] != self.input_length:
            raise ValueError(f"Bad shape: ({X.shape[1]} != {self.input_length})")

        # Mandatory copy in order not to modify the original !
        if isinstance(X, pd.DataFrame):
            return X.copy(deep=True)
        else:
            return X.copy()

    def fit(self, X: Union[np.ndarray, pd.DataFrame], y: Union[np.ndarray, pd.Series, pd.DataFrame]) -> Any:
        '''Fit transformer

        Args:
            X (np.ndarray or pd.DataFrame): Array-like, shape = [n_samples, n_features]
            y (np.ndarray or pd.Series or pd.DataFrame): Array-like, shape = [n_samples, n_targets]
        Returns:
            self
        '''
        raise NotImplementedError("'fit' needs to be overridden")

    def transform(self, X: Union[np.ndarray, pd.DataFrame]) -> np.ndarray:
        '''Transforms X

        Args:
            X (np.ndarray or pd.DataFrame): Array-like, shape = [n_samples, n_features]
        Returns:
            Transformed X
        '''
        raise NotImplementedError("'transform' needs to be overridden")

    def fit_transform(self, X: Union[np.ndarray, pd.DataFrame], y: Union[np.ndarray, pd.Series, pd.DataFrame, None] = None) -> np.ndarray:
        '''Applies both fit & transform.

        Args:
            X (np.ndarray or pd.DataFrame): Shape = [n_samples, n_features]
        Kwargs:
            y: Array-like, shape = [n_samples]
        Returns:
            X_out: Array-like, shape [n_samples, n_features]
                        Transformed input.
        '''
        self.fit(X, y)
        return self.transform(X)


class AutoLogTransform(Estimator):
    '''Automatically applies a log transformation on numerical data if the distribution of the variables
       is skewed (abs(skew) > min_skewness) and if there is an amplitude superior to min_amplitude between
       the 10th and 90th percentiles

    WARNING: YOUR DATA MUST BE POSITIVE IN ORDER FOR EVERYTHING TO WORK CORRECTLY
    '''

    def __init__(self, min_skewness: float = 2, min_amplitude: float = 10E3) -> None:
        '''Initialization of the class

        Kwargs:
            min_skewness (float): Absolute value of the required skewness to apply a log transformation
            min_amplitude (float): Minimal value of the amplitude between the 10th and 90th percentiles
                                   required to apply a log transformation
        '''

        super().__init__(None)
        # Set attributes
        self.min_skewness = min_skewness
        self.min_amplitude = min_amplitude

        # Columns on which to apply the transformation
        # Set on fit
        # Warning: sklearn does not support columns name, so we can only use indexes
        # Hence, X input must expose same columns order (this won't be checked)
        self.applicable_columns_index: Optional[List[Any]] = None

    def fit(self, X: Union[np.ndarray, pd.DataFrame], y: Any = None) -> Any:
        '''Fit transformer

        Args:
            X (np.ndarray or pd.DataFrame): Array-like, shape = [n_samples, n_features]
        Kwargs:
            y (None): Not used here
        Returns:
            self
        '''
        X = self._validate_input(X)
        # If x is a numpy array, casts it in pd.DataFrame
        if isinstance(X, np.ndarray):
            X = pd.DataFrame(X)
        # Otherwise, we reset the columns name because sklearn can't manage them
        else:
            X = X.rename(columns={col: i for i, col in enumerate(X.columns)})
        self.input_length = X.shape[1]

        # Get applicable columns
        skew = X.skew()
        candidates = list(skew[abs(skew) > self.min_skewness].index)
        if len(candidates) > 0:
            q10 = X.iloc[:, candidates].quantile(q=0.1)
            q90 = X.iloc[:, candidates].quantile(q=0.9)
            amp = q90 - q10
            # Update applicable_columns_index
            self.applicable_columns_index = list(amp[amp > self.min_amplitude].index)

        self.fitted_ = True
        return self

    def transform(self, X: Union[np.ndarray, pd.DataFrame]) -> np.ndarray:
        '''Transforms X - apply log on applicable columns

        Args:
            X (np.ndarray or pd.DataFrame): Array-like, shape = [n_samples, n_features]
        Returns:
            X_out: Array-like, shape [n_samples, n_features]
                        Transformed input.
        '''
        # Validate input
        check_is_fitted(self, 'fitted_')
        X = self._validate_input(X)

        # If x is a numpy array, casts it in pd.DataFrame
        if isinstance(X, np.ndarray):
            X = pd.DataFrame(X)

        # Log transformation on the applicable columns
        if len(self.applicable_columns_index) > 0:
            X.iloc[:, self.applicable_columns_index] = np.log(X.iloc[:, self.applicable_columns_index])

        # Compatibility -> returns numpy array
        return X.to_numpy()


class ThresholdingTransform(Estimator):
    '''Applies a threshold on columns.
    If min and max values are given, the threshold is manual; otherwise it is statistical.
    '''

    def __init__(self, thresholds: List[Tuple], quantiles: tuple = (0.05, 0.95)) -> None:
        '''Initialization of the class

        Args:
            tresholds (list of tuple): Each tuple contains (min_val, max_val).
        Kwargs:
            quantiles (tuple): Tuple containing (quantile_min, quantile_max)
        Raises:
            ValueError: If quantiles values are not between 0 and 1 and if quantiles[0] >= quantiles[1]
        '''
        if not 0 < quantiles[0] < 1 or not 0 < quantiles[1] < 1 or not quantiles[0] < quantiles[1]:
            raise ValueError(f"The values contained in quantiles should verify quantile_min < quantile_max and both > 0 and < 1. quantiles = {quantiles} is not supported.")

        super().__init__(len(thresholds))

        # Set attributes
        self.thresholds = thresholds
        self.fitted_thresholds: List[tuple] = []
        self.quantiles = quantiles

    def fit(self, X: Union[np.ndarray, pd.DataFrame], y: Any = None) -> Any:
        '''Fits the ThresholdingTransform on X.

        Args:
            X (np.ndarray or pd.DataFrame): Shape [n_samples, n_features]
        Kwargs:
            y (None): Not used here.
        Returns:
            self: ThresholdingTransform
        '''
        X = self._validate_input(X)
        # If X is a numpy array, casts it as a pd.DataFrame
        if isinstance(X, np.ndarray):
            X = pd.DataFrame(X)

        # Fits column one by one
        for col_index, item in enumerate(self.thresholds):
            val_min, val_max = item
            if val_min is None:
                val_min = X.iloc[:, col_index].quantile(q=self.quantiles[0])
            if val_max is None:
                val_max = X.iloc[:, col_index].quantile(q=self.quantiles[1])
            self.fitted_thresholds.append((col_index, val_min, val_max))

        self.fitted_ = True
        return self

    def transform(self, X: Union[np.ndarray, pd.DataFrame]) -> np.ndarray:
        '''Impute all missing values in X.

        Args:
            X (np.ndarray or pd.DataFrame): Shape (n_samples, n_features)
                The input data to complete.
        '''
        check_is_fitted(self, 'fitted_')
        X = self._validate_input(X)
        # If X is a numpy array, casts it to pd.DataFrame
        if isinstance(X, np.ndarray):
            X = pd.DataFrame(X)

        for col_index, val_min, val_max in self.fitted_thresholds:
            X.iloc[:, col_index][X.iloc[:, col_index] < val_min] = val_min
            X.iloc[:, col_index][X.iloc[:, col_index] > val_max] = val_max

        return X.to_numpy() # Compatibility -> return a numpy array


class AutoBinner(Estimator):
    '''Automatically creates a "other" category when the categories are heavily unbalanced
    /!\ Replaces the values of some categories /!\
    '''

    def __init__(self, strategy: str = "auto", min_cat_count: int = 3, threshold: float = 0.05) -> None:
        '''Initialization of the class

        Kwargs:
            strategy (str): 'auto' or 'threshold'
                - 'auto': Aggregates all categories as long as their cumulated frequence is less than threshold
                - 'threshold': Aggregates all category whose frequence is less than threshold
            min_cat_count (int): Minimal number of category to keep
            threshold (float): The threshold to consider
        Raises:
            ValueError: The object strategy must be in the list of allowed strategies
            ValueError: The object min_cat_count must be non negative
            ValueError: The object threshold must be in ]0,1[
        '''
        super().__init__(None)

        allowed_strategies = ["threshold", "auto"]
        self.strategy = strategy
        if self.strategy not in allowed_strategies:
            raise ValueError(f"Can only use these strategies: {allowed_strategies}. " +\
                             f"Got strategy={strategy}")
        if min_cat_count < 0:
            raise ValueError("min_cat_count must be non negative")
        if not (0 < threshold < 1):
            raise ValueError(f"threshold must be in ]0,1[, not {threshold}")

        # Set attributes
        self.min_cat_count = min_cat_count
        self.threshold = threshold
        self.kept_cat_by_index: Dict[int, list] = {}

    def fit(self, X: Union[np.ndarray, pd.DataFrame], y: Any = None) -> Any:
        '''Fit the AutoBinner on X.

        Args:
            X (np.ndarray or pd.DataFrame): Shape (n_samples, n_features)
        Kwargs:
            y: Not used here
        Returns:
            self: AutoBinner
        '''
        X = self._validate_input(X)
        # If x is a numpy array, casts it in pd.DataFrame
        if isinstance(X, np.ndarray):
            X = pd.DataFrame(X)

        self.input_length = X.shape[1]
        # Fits column one by one
        for col_index in range(self.input_length):
            # Get col serie
            X_tmp_ser = X.iloc[:, col_index]
            # Get unique vals
            unique_cat = list(X_tmp_ser.unique())
            # If less vals than min threshold, set this column allowed values with all uniques values
            if len(unique_cat) <= self.min_cat_count:
                self.kept_cat_by_index[col_index] = unique_cat
                continue

            # If more vals than min threshold, keep values based on strategy
            table = X_tmp_ser.value_counts() / X_tmp_ser.count()
            table = table.sort_values()
            if self.strategy == 'auto':
                table = np.cumsum(table)
            # If only one category is less than the threshold, we do not need to transform it
            if table[1] > self.threshold:
                self.kept_cat_by_index[col_index] = unique_cat
                continue
            # Otherwise, we get rid of the superfluous categories
            else:
                to_remove = list(table[table<self.threshold].index)
                for item in to_remove:
                    unique_cat.remove(item)
                self.kept_cat_by_index[col_index] = unique_cat

        self.fitted_ = True
        return self

    def transform(self, X: Union[np.ndarray, pd.DataFrame]) -> np.ndarray:
        '''Imputes all missing values in X.

        Args:
            X (np.ndarray or pd.DataFrame): Shape (n_samples, n_features)
                The input data to complete.
        '''
        check_is_fitted(self, 'fitted_')
        X = self._validate_input(X)
        # If x is a numpy array, casts it in pd.DataFrame
        if isinstance(X, np.ndarray):
            X = pd.DataFrame(X)

        for col_index in range(self.input_length):
            X.iloc[:, col_index] = X.iloc[:, col_index].apply(lambda x: x if x in self.kept_cat_by_index[col_index] else 'other_')

        return X.to_numpy() # Compatibility, returns a numpy array


class EmbeddingTransformer(Estimator):
    '''Constructs a transformer that apply an embedding mapping to Categorical columns'''

    def __init__(self, embedding: Union[str, dict], none_strategy: str = 'zeros') -> None:
        '''Initialization of the class

        Args:
            embedding (str or dict): Embedding to use
                - If dict -> ok, ready to go
                - If str -> path to the file to load (json)
        Kwargs:
            none_strategy (str): Strategy to fill elements not in embedding
                - Zeros: only 0s
        Raises:
            ValueError: If strategy is not in the allowed strategies
            ValueError: If the embedding is of type str but does not end in .json
            FileNotFoundError: If the path to the embedding does not exist
        '''
        super().__init__(None)

        # Check none strategy
        allowed_strategies = ["zeros"]
        self.none_strategy = none_strategy
        if self.none_strategy not in allowed_strategies:
           raise ValueError(f"Can only use these strategies: {allowed_strategies}, got strategy={self.none_strategy}")

        # If str, loads the embedding
        if isinstance(embedding, str):
            if not embedding.endswith('.json'):
                raise ValueError(f"The file {embedding} must be a .json file")
            if not os.path.exists(embedding):
                raise FileNotFoundError(f"The file {embedding} does not exist")
            with open(embedding, 'r', encoding='{{default_encoding}}') as f:
                self.embedding = json.load(f)
        else:
            self.embedding = embedding

        # Get embedding size
        self.embedding_size = len(self.embedding[list(self.embedding.keys())[0]])
        # Other params
        self.n_missed = 0

    def fit(self, X: Union[np.ndarray, pd.DataFrame], y: Any = None) -> Any:
        '''Fit transformer

        Args:
            X (np.ndarray or pd.DataFrame): Shape (n_samples, n_features)
        Kwargs:
            y: Not used here
        Returns
            self (EmbeddingTransformer)
        '''
        X = self._validate_input(X)
        # If x is a numpy array, casts it in pd.DataFrame
        if isinstance(X, np.ndarray):
            X = pd.DataFrame(X)

        self.input_length = X.shape[1]

        # Nothing to do

        self.fitted_ = True
        return self

    def transform(self, X: Union[np.ndarray, pd.DataFrame]) -> np.ndarray:
        '''Transform X - embedding mapping

        Args:
            X (np.ndarray or pd.DataFrame): Shape (n_samples, n_features)
        Raises:
            ValueError: If there are missing columns
        Returns:
            X_out (np.ndarray): Shape (n_samples, n_features) transformed input.
        '''
        X = self._validate_input(X)
        # If x is a numpy array, casts it in pd.DataFrame
        if isinstance(X, np.ndarray):
            X = pd.DataFrame(X)

        n_rows = X.shape[0]

        # Apply mapping
        new_df = pd.DataFrame()
        for col in X.columns:
            self.n_missed = 0  # Counts the number of missing elements in the embedding
            tmp_serie = X[col].apply(self.apply_embedding)  # Updates self.n_missed
            new_df = pd.concat([new_df, pd.DataFrame(tmp_serie.to_list())], axis=1)
            perc_missed = self.n_missed / n_rows * 100
            if perc_missed != 0:
                logger.warning(f"Warning, {self.n_missed} ({perc_missed} %) missing elements in the embedding for column {col}")

        return new_df.to_numpy() # Compatibility, returns a numpy array

    def apply_embedding(self, content) -> list:
        '''Applies embedding mapping

        Args:
            content: Content on which to apply embedding mapping
        Raises:
            ValueError: If the strategy is not recognized
        Returns:
            list: Applied embedding
        '''
        if content in self.embedding.keys():
            return self.embedding[content]
        else:
            self.n_missed += 1
            if self.none_strategy == 'zeros':
                return [0] * self.embedding_size
            else:
                raise ValueError(f"Strategy {self.none_strategy} not recognized")

    def get_feature_names(self, features_in: list, *args, **kwargs) -> np.ndarray:
        '''Returns feature names for output features.

        Args:
            features_in (list): list of features

        Returns:
            output_feature_names: ndarray of shape (n_output_features,)
                Array of feature names.
        '''
        check_is_fitted(self, 'fitted_')
        new_features = [f"emb_{feat}_{i}" for feat in features_in for i in range(self.embedding_size)]
        return np.array(new_features, dtype=object)


if __name__ == '__main__':
    logger.error("This script is not stand alone but belongs to a package that has to be imported.")



#!/usr/bin/env python3
# Copyright (C) <2018-2021>  <Agence Data Services, DSI Pôle Emploi>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

# Libs unittest
import unittest
from unittest.mock import Mock
from unittest.mock import patch

# Utils libs
import os
import shutil
import pickle
import numpy as np
import pandas as pd

from {{package_name}} import utils
from {{package_name}}.preprocessing import preprocess
from {{package_name}}.models_training import utils_models
from {{package_name}}.models_training import model_tfidf_svm, model_embedding_lstm

# Disable logging
import logging
logging.disable(logging.CRITICAL)


def remove_dir(path):
    if os.path.isdir(path): shutil.rmtree(path)


class UtilsModelsTests(unittest.TestCase):
    '''Main class to test all functions in utils_models.py'''

    # We avoid tqdm prints
    pd.Series.progress_apply = pd.Series.apply

    def setUp(self):
        '''SetUp fonction'''
        # Change directory to script directory
        abspath = os.path.abspath(__file__)
        dname = os.path.dirname(abspath)
        os.chdir(dname)

    def tearDown(self):
        '''Cleaning fonction -> we delete the mock embedding'''
        data_path = utils.get_data_path()
        fake_path = os.path.join(data_path, 'fake_embedding.pkl')
        if os.path.exists(fake_path):
            os.remove(fake_path)

    def test01_get_embedding(self):
        '''Test of the function utils_models.get_embedding'''
        # Create a mock embedding
        data_path = utils.get_data_path()
        # Check if data folder exists
        if not os.path.exists(data_path):
            os.mkdir(data_path)
        fake_embedding = {'toto': [0.25, 0.30], 'titi': [0.85, 0.12]}
        fake_path = os.path.join(data_path, 'fake_embedding.pkl')
        if os.path.exists(fake_path):
            os.remove(fake_path)
        with open(fake_path, 'wb') as f:
            pickle.dump(fake_embedding, f, pickle.HIGHEST_PROTOCOL)

        # Nominal case
        reloaded_embedding = utils_models.get_embedding('fake_embedding.pkl')
        self.assertEqual(reloaded_embedding, fake_embedding)

    def test02_normal_split(self):
        '''Test of the function {{package_name}}.models_training.utils_models.normal_split'''
        # Valids to test
        input_test = pd.DataFrame({'col1': ['a', 'b', 'c', 'd', 'e', 'f', 'g', 'h', 'i', 'j']})
        test_size = 0.2
        train, test = utils_models.normal_split(input_test, test_size=test_size)
        self.assertEqual(train.shape[0], input_test.shape[0] * (1 - test_size))
        self.assertEqual(test.shape[0], input_test.shape[0] * test_size)

        # Check inputs
        with self.assertRaises(ValueError):
            utils_models.normal_split(input_test, test_size=1.2)
        with self.assertRaises(ValueError):
            utils_models.normal_split(input_test, test_size=-0.2)

    def test03_stratified_split(self):
        '''Test of the function {{package_name}}.models_training.utils_models.stratified_split'''
        # Valids to test
        input_test = pd.DataFrame({'col1': ['a', 'b', 'c', 'd', 'e', 'f', 'g', 'h', 'i', 'j', 'k'],
                                   'col2': [0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 2]})
        test_size = 0.5
        col = 'col2'
        train, test = utils_models.stratified_split(input_test, col, test_size=test_size)
        # -1 because the '2' is removed (not enough values)
        self.assertEqual(train.shape[0], (input_test.shape[0] - 1) * (1 - test_size))
        self.assertEqual(test.shape[0], (input_test.shape[0] - 1) * test_size)
        self.assertEqual(train[train[col] == 0].shape[0], input_test[input_test[col] == 0].shape[0] * (1 - test_size))
        self.assertEqual(test[test[col] == 0].shape[0], input_test[input_test[col] == 0].shape[0] * test_size)
        self.assertEqual(train[train[col] == 2].shape[0], 0)
        self.assertEqual(test[test[col] == 2].shape[0], 0)

        # Check the input(s) type(s)
        with self.assertRaises(ValueError):
            utils_models.stratified_split(input_test, col, test_size=1.2)
        with self.assertRaises(ValueError):
            utils_models.stratified_split(input_test, col, test_size=-0.2)

    def test04_hierarchical_split(self):
        '''Test of the function {{package_name}}.models_training.utils_models.hierarchical_split'''
        # Valids to test
        duplicate_val = 'test1'
        input_test = pd.DataFrame({'col1': ['x', 'a', 'b', 'c', 'x', 'd', 'e', 'f', 'g', 'h', 'x', 'x', 'x', 'x', 'i'],
                                   'col2': [duplicate_val, 'a', 'b', 'c', duplicate_val, 'd', 'e', 'f', 'g', 'h', duplicate_val, duplicate_val, duplicate_val, duplicate_val, 'i']})
        test_size = 0.2
        col = 'col2'
        train, test = utils_models.hierarchical_split(input_test, col, test_size=test_size)
        # We can't really test the dataframe size here (depends on where the duplicate value is sent)
        # Instead, we test the nb of unique values
        self.assertEqual(len(train[col].unique()), len(input_test[col].unique()) * (1 - test_size))
        self.assertEqual(len(test[col].unique()), len(input_test[col].unique()) * test_size)
        # Check all 6 duplicates value in the same dataframe
        self.assertTrue((train[train[col] == duplicate_val].shape[0] == 6 and test[test[col] == duplicate_val].shape[0] == 0)
                        or (train[train[col] == duplicate_val].shape[0] == 0 and test[test[col] == duplicate_val].shape[0] == 6))

        # Check the input(s) type(s)
        with self.assertRaises(ValueError):
            utils_models.hierarchical_split(input_test, col, test_size=1.2)
        with self.assertRaises(ValueError):
            utils_models.hierarchical_split(input_test, col, test_size=-0.2)

    def test05_remove_small_classes(self):
        '''Test of the function {{package_name}}.models_training.utils_models.remove_small_classes'''
        # Valids to test
        input_test = pd.DataFrame({'col1': ['a', 'b', 'c', 'd', 'e', 'f', 'g', 'h', 'i', 'j', 'k'],
                                   'col2': [0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 2]})
        test_size = 0.2
        col = 'col2'

        result_df = utils_models.remove_small_classes(input_test, col, min_rows=2)
        self.assertEqual(result_df[result_df[col] == 0].shape[0], input_test[input_test[col] == 0].shape[0])
        self.assertEqual(result_df[result_df[col] == 1].shape[0], input_test[input_test[col] == 1].shape[0])
        self.assertEqual(result_df[result_df[col] == 2].shape[0], 0)

        result_df = utils_models.remove_small_classes(input_test, col, min_rows=5)
        self.assertEqual(result_df[result_df[col] == 0].shape[0], 0)
        self.assertEqual(result_df[result_df[col] == 1].shape[0], input_test[input_test[col] == 1].shape[0])
        self.assertEqual(result_df[result_df[col] == 2].shape[0], 0)

        # Check the input(s) type(s)
        with self.assertRaises(ValueError):
            utils_models.remove_small_classes(input_test, col, min_rows=0)

    def test06_display_train_test_shape(self):
        '''Test of the function {{package_name}}.models_training.utils_models.display_train_test_shape'''
        # Valids to test
        df = pd.DataFrame({'col1': ['a', 'b', 'c', 'd', 'e', 'f', 'g', 'h', 'i', 'j', 'k'],
                           'col2': [0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 2]})

        # Nominal case
        utils_models.display_train_test_shape(df, df)
        utils_models.display_train_test_shape(df, df, df_shape=10)

    def test07_preprocess_model_multilabel(self):
        '''Test of the function {{package_name}}.models_training.utils_models.preprocess_model_multilabel'''
        # Creation of a dataset
        df = pd.DataFrame({'x_col': ['test', 'toto', 'titi'], 'y_col': [(), ('x1', 'x2'), ('x3', 'x4', 'x1')]})
        df_expected = pd.DataFrame({'x_col': ['test', 'toto', 'titi'], 'y_col': [(), ('x1', 'x2'), ('x3', 'x4', 'x1')],
                                    'x1': [0, 1, 1], 'x2': [0, 1, 0], 'x3': [0, 0, 1], 'x4': [0, 0, 1]})
        subset_classes = ['x2', 'x4']
        df_subset_expected = pd.DataFrame({'x_col': ['test', 'toto', 'titi'], 'y_col': [(), ('x1', 'x2'), ('x3', 'x4', 'x1')],
                                           'x2': [0, 1, 0], 'x4': [0, 0, 1]})

        # Nominal case
        df_mlb, classes = utils_models.preprocess_model_multilabel(df, 'y_col')
        self.assertEqual(sorted(classes), ['x1', 'x2', 'x3', 'x4'])
        pd.testing.assert_frame_equal(df_mlb, df_expected, check_dtype=False)

        # Test argument classes
        df_mlb, classes = utils_models.preprocess_model_multilabel(df, 'y_col', classes=subset_classes)
        self.assertEqual(sorted(classes), sorted(subset_classes))
        pd.testing.assert_frame_equal(df_mlb, df_subset_expected, check_dtype=False)

    def test08_load_model(self):
        '''Test of the function {{package_name}}.models_training.utils_models.load_model'''
        # Creation fake model
        x_train = ['test', 'test', 'test', 'toto', 'toto', 'toto']
        y_train = [0, 0, 0, 1, 1, 1]
        model_dir = os.path.join(utils.get_models_path(), 'test_model')
        remove_dir(model_dir)
        model_name = 'test_model_name'
        model = model_tfidf_svm.ModelTfidfSvm(model_dir=model_dir, model_name=model_name)
        model.fit(x_train, y_train)
        model.save()

        # Reload
        new_model, new_config = utils_models.load_model(model_dir='test_model')
        # We perform some tests
        self.assertEqual(new_config['model_name'], model_name)
        self.assertEqual(new_model.model_name, model_name)
        self.assertEqual(list(new_model.predict(['test', 'toto', 'a'])), list(model.predict(['test', 'toto', 'a'])))

        # Same thing but with a path
        new_model, new_config = utils_models.load_model(model_dir=model_dir, is_path=True)
        # We perform some tests
        self.assertEqual(new_config['model_name'], model_name)
        self.assertEqual(new_model.model_name, model_name)
        self.assertEqual(list(new_model.predict(['test', 'toto', 'a'])), list(model.predict(['test', 'toto', 'a'])))
        remove_dir(model_dir)

        # We do the same thing on a keras model
        data_path = utils.get_data_path()
        # Check if data folder exists
        if not os.path.exists(data_path):
            os.mkdir(data_path)
        fake_embedding = {'test': [0.8, 0.2], 'toto': [0.1, 0.4]}
        fake_path = os.path.join(data_path, 'fake_embedding.pkl')
        if os.path.exists(fake_path):
            os.remove(fake_path)
        with open(fake_path, 'wb') as f:
            pickle.dump(fake_embedding, f, pickle.HIGHEST_PROTOCOL)
        x_train = ['test', 'test', 'test', 'toto', 'toto', 'toto'] * 10
        y_train = [0, 0, 0, 1, 1, 1] * 10
        model_dir = os.path.join(utils.get_models_path(), 'test_model_dl')
        remove_dir(model_dir)
        model_name = 'test_model_dl_name'
        epochs = 2
        batch_size = 8
        embedding_name = 'fake_embedding.pkl'
        model = model_embedding_lstm.ModelEmbeddingLstm(model_dir=model_dir, model_name=model_name,
                                                        epochs=epochs, batch_size=batch_size,
                                                        embedding_name=embedding_name)
        model.fit(x_train, y_train)
        model.save()

        # Reload
        new_model, new_config = utils_models.load_model(model_dir='test_model_dl')
        # We perform some tests
        self.assertEqual(new_config['model_name'], model_name)
        self.assertEqual(new_config['epochs'], epochs)
        self.assertEqual(new_config['batch_size'], batch_size)
        self.assertEqual(new_config['embedding_name'], embedding_name)
        self.assertEqual(new_model.model_name, model_name)
        self.assertEqual(new_model.epochs, epochs)
        self.assertEqual(new_model.batch_size, batch_size)
        self.assertEqual(new_model.embedding_name, embedding_name)
        self.assertEqual(list(new_model.predict(['test', 'toto', 'a'])), list(model.predict(['test', 'toto', 'a'])))
        self.assertEqual([list(_) for _ in new_model.predict_proba(['test', 'toto', 'a'])], [list(_) for _ in model.predict_proba(['test', 'toto', 'a'])])
        remove_dir(model_dir)

        # TODO: test pytorch models ?

        # Check errors
        with self.assertRaises(FileNotFoundError):
            utils_models.load_model(model_dir='tototo')
        with self.assertRaises(FileNotFoundError):
            utils_models.load_model(model_dir='./tototo', is_path=True)

    def test09_predict(self):
        '''Test of the function demo_torch.models_training.utils_models.predict'''
        # Data
        x_train_base = ['Ceci est un test', 'Ceci est un test', 'Ceci est un test', 'Test "deux" !', 'Toto "deux" !', 'Test "deux" !', 'deux'] * 10
        y_train_mono = ['test', 'test', 'test', 'deux', 'deux', 'deux', 'deux'] * 10
        y_train_multi = pd.DataFrame({'test': [1, 1, 1, 1, 0, 1, 0] * 10, 'toto': [0, 0, 0, 0, 1, 0, 0] * 10})
        preprocessor = preprocess.get_preprocessor('preprocess_P1')
        x_train = preprocessor(x_train_base)

        # Nominal case - mono-label
        model_dir = os.path.join(utils.get_models_path(), 'test_model3')
        remove_dir(model_dir)
        model_name = 'test_model3_name'
        model = model_tfidf_svm.ModelTfidfSvm(model_dir=model_dir, model_name=model_name)
        model.fit(x_train, y_train_mono)
        model.save()
        model, model_conf = utils_models.load_model('test_model3')
        self.assertEqual(utils_models.predict('Ceci est un test', model, model_conf), 'test')
        self.assertEqual(utils_models.predict('Test "deux" !', model, model_conf), 'deux')
        remove_dir(model_dir)

        # Nominal case - multi-labels
        model_dir = os.path.join(utils.get_models_path(), 'test_model3')
        remove_dir(model_dir)
        model_name = 'test_model3_name'
        model = model_tfidf_svm.ModelTfidfSvm(model_dir=model_dir, model_name=model_name, multi_label=True)
        model.fit(x_train, y_train_multi)
        model.save()
        model, model_conf = utils_models.load_model('test_model3')
        self.assertEqual(utils_models.predict('Ceci est un test', model, model_conf), ('test',))
        self.assertEqual(utils_models.predict('Toto "deux" !', model, model_conf), ('toto',))
        remove_dir(model_dir)

    def test10_predict_with_proba(self):
        '''Test of the function demo_torch.models_training.utils_models.predict_with_proba'''
        # Data
        x_train_base = ['Ceci est un test', 'Ceci est un test', 'Ceci est un test', 'Test "deux" !', 'Toto "deux" !', 'Test "deux" !', 'deux'] * 10
        y_train_mono = ['test', 'test', 'test', 'deux', 'deux', 'deux', 'deux'] * 10
        y_train_multi = pd.DataFrame({'test': [1, 1, 1, 1, 0, 1, 0] * 10, 'toto': [0, 0, 0, 0, 1, 0, 0] * 10})
        preprocessor = preprocess.get_preprocessor('preprocess_P1')
        x_train = preprocessor(x_train_base)

        # Nominal case - mono-label
        model_dir = os.path.join(utils.get_models_path(), 'test_model4')
        remove_dir(model_dir)
        model_name = 'test_model4_name'
        model = model_tfidf_svm.ModelTfidfSvm(model_dir=model_dir, model_name=model_name)
        model.fit(x_train, y_train_mono)
        model.save()
        model, model_conf = utils_models.load_model('test_model4')
        pred, proba = utils_models.predict_with_proba('Ceci est un test', model, model_conf)
        self.assertEqual(pred, 'test')
        self.assertEqual(proba, 1.0) # 1.0 because svm
        pred, proba = utils_models.predict_with_proba('Test "deux" !', model, model_conf)
        self.assertEqual(pred, 'deux')
        self.assertEqual(proba, 1.0) # 1.0 because svm
        remove_dir(model_dir)

        # Nominal case - multi-labels
        model_dir = os.path.join(utils.get_models_path(), 'test_model4')
        remove_dir(model_dir)
        model_name = 'test_model4_name'
        model = model_tfidf_svm.ModelTfidfSvm(model_dir=model_dir, model_name=model_name, multi_label=True)
        model.fit(x_train, y_train_multi)
        model.save()
        model, model_conf = utils_models.load_model('test_model4')
        pred, proba = utils_models.predict_with_proba('Ceci est un test', model, model_conf)
        self.assertEqual(pred, ('test',))
        self.assertEqual(proba, (1.0,)) # 1.0 because svm
        pred, proba = utils_models.predict_with_proba('Toto "deux" !', model, model_conf)
        self.assertEqual(pred, ('toto',))
        self.assertEqual(proba, (1.0,)) # 1.0 because svm
        remove_dir(model_dir)

    def test11_search_hp_cv(self):
        '''Test of the function {{package_name}}.models_training.utils_models.search_hp_cv'''
        # Definition of the variables for the nominal case
        x_train = ['Ceci est un test', 'Ceci est un test', 'Ceci est un test', 'Test "deux" !', 'Toto "deux" !', 'Test "deux" !', 'deux'] * 10
        y_train_mono = ['test', 'test', 'test', 'deux', 'deux', 'deux', 'deux'] * 10
        y_train_multi = pd.DataFrame({'test': [1, 1, 1, 1, 0, 1, 0] * 10, 'toto': [0, 0, 0, 0, 1, 0, 0] * 10})
        model_cls = model_tfidf_svm.ModelTfidfSvm
        model_params_mono = {'multi_label': False}
        model_params_multi = {'multi_label': True}
        hp_params = {'tfidf_params': [{'analyzer': 'word', 'ngram_range': (1, 2), "max_df": 1.0, 'min_df': 0.0}, {'analyzer': 'word', 'ngram_range': (1, 3), "max_df": 1.0, 'min_df': 0.0}], 'svc_params': [{'C': 1.0}, {'C': 2.0}]}
        kwargs_fit_mono = {'x_train':x_train, 'y_train': y_train_mono}
        kwargs_fit_multi = {'x_train':x_train, 'y_train': y_train_multi}

        # Nominal case
        n_splits = 5
        model = utils_models.search_hp_cv(model_cls, model_params_mono, hp_params, "accuracy", kwargs_fit_mono, n_splits=n_splits)
        self.assertFalse(model.trained)
        self.assertEqual(model.nb_fit, 0)
        csv_path = os.path.join(model.model_dir, f"hyper_params_results.csv")
        json_path = os.path.join(model.model_dir, f"hyper_params_tested.json")
        self.assertTrue(os.path.exists(csv_path))
        self.assertTrue(os.path.exists(json_path))
        with open(csv_path, 'r', encoding="{{default_encoding}}") as f:
            self.assertEqual(sum(1 for line in f), n_splits * len(hp_params) + 1)
        remove_dir(model.model_dir)

        # Multi-labels case
        n_splits = 3
        model = utils_models.search_hp_cv(model_cls, model_params_multi, hp_params, "f1", kwargs_fit_multi, n_splits=n_splits)
        self.assertFalse(model.trained)
        self.assertEqual(model.nb_fit, 0)
        csv_path = os.path.join(model.model_dir, f"hyper_params_results.csv")
        json_path = os.path.join(model.model_dir, f"hyper_params_tested.json")
        self.assertTrue(os.path.exists(csv_path))
        self.assertTrue(os.path.exists(json_path))
        with open(csv_path, 'r', encoding="{{default_encoding}}") as f:
            self.assertEqual(sum(1 for line in f), n_splits * len(hp_params) + 1)
        remove_dir(model.model_dir)

        # Check the various scoring functions
        n_splits = 2
        model = utils_models.search_hp_cv(model_cls, model_params_mono, hp_params, "precision", kwargs_fit_mono, n_splits=n_splits)
        self.assertFalse(model.trained)
        self.assertEqual(model.nb_fit, 0)
        csv_path = os.path.join(model.model_dir, f"hyper_params_results.csv")
        json_path = os.path.join(model.model_dir, f"hyper_params_tested.json")
        self.assertTrue(os.path.exists(csv_path))
        self.assertTrue(os.path.exists(json_path))
        with open(csv_path, 'r', encoding="{{default_encoding}}") as f:
            self.assertEqual(sum(1 for line in f), n_splits * len(hp_params) + 1)
        remove_dir(model.model_dir)

        n_splits = 2
        model = utils_models.search_hp_cv(model_cls, model_params_mono, hp_params, "recall", kwargs_fit_mono, n_splits=n_splits)
        self.assertFalse(model.trained)
        self.assertEqual(model.nb_fit, 0)
        csv_path = os.path.join(model.model_dir, f"hyper_params_results.csv")
        json_path = os.path.join(model.model_dir, f"hyper_params_tested.json")
        self.assertTrue(os.path.exists(csv_path))
        self.assertTrue(os.path.exists(json_path))
        with open(csv_path, 'r', encoding="{{default_encoding}}") as f:
            self.assertEqual(sum(1 for line in f), n_splits * len(hp_params) + 1)
        remove_dir(model.model_dir)

        # Custom scoring function
        def custom_func(test_dict: dict):
            return (test_dict['Precision'] + test_dict['Recall']) / 2
        n_splits = 2
        model = utils_models.search_hp_cv(model_cls, model_params_mono, hp_params, custom_func, kwargs_fit_mono, n_splits=n_splits)
        self.assertFalse(model.trained)
        self.assertEqual(model.nb_fit, 0)
        csv_path = os.path.join(model.model_dir, f"hyper_params_results.csv")
        json_path = os.path.join(model.model_dir, f"hyper_params_tested.json")
        self.assertTrue(os.path.exists(csv_path))
        self.assertTrue(os.path.exists(json_path))
        with open(csv_path, 'r', encoding="{{default_encoding}}") as f:
            self.assertEqual(sum(1 for line in f), n_splits * len(hp_params) + 1)
        remove_dir(model.model_dir)

        # Check errors
        with self.assertRaises(TypeError):
            model = utils_models.search_hp_cv(model_cls, model_params_mono, hp_params, 5, kwargs_fit_mono, n_splits=n_splits)
        with self.assertRaises(ValueError):
            model = utils_models.search_hp_cv(model_cls, model_params_mono, hp_params, 'toto', kwargs_fit_mono, n_splits=n_splits)
        with self.assertRaises(ValueError):
            model = utils_models.search_hp_cv(model_cls, {'toto': True}, hp_params, 'accuracy', kwargs_fit_mono, n_splits=n_splits)
        with self.assertRaises(ValueError):
            model = utils_models.search_hp_cv(model_cls, model_params_mono, hp_params, 'accuracy', {'toto': True, 'y_train': y_train_mono}, n_splits=n_splits)
        with self.assertRaises(ValueError):
            model = utils_models.search_hp_cv(model_cls, model_params_mono, hp_params, 'accuracy', {'toto': True, 'x_train': x_train}, n_splits=n_splits)
        with self.assertRaises(ValueError):
            model = utils_models.search_hp_cv(model_cls, model_params_mono, hp_params, 'accuracy', {'toto': True, 'x_train': x_train}, n_splits=n_splits)
        with self.assertRaises(ValueError):
            model = utils_models.search_hp_cv(model_cls, {**model_params_mono, **{'toto': True}}, {**hp_params, **{'toto': [False]}}, 'accuracy', kwargs_fit_mono, n_splits=n_splits)
        with self.assertRaises(ValueError):
            model = utils_models.search_hp_cv(model_cls, model_params_mono, {'toto': [1, 2], 'titi': [3]}, 'accuracy', kwargs_fit_mono, n_splits=n_splits)
        with self.assertRaises(ValueError):
            model = utils_models.search_hp_cv(model_cls, model_params_mono, hp_params, 'accuracy', kwargs_fit_mono, n_splits=1)


# Perform tests
if __name__ == '__main__':
    # Start tests
    unittest.main()



#!/usr/bin/env python3

## Data preprocessing
# Copyright (C) <2018-2022>  <Agence Data Services, DSI Pôle Emploi>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# Ex: python 1_preprocess_data.py -f original_dataset.csv train.csv test.csv --encoding utf-8 --sep ; --input_col text


import os
import gc
import logging
import argparse
import pandas as pd
from pathlib import Path
from typing import Union, List

from {{package_name}} import utils
from {{package_name}}.preprocessing import preprocess

# Get logger
logger = logging.getLogger('{{package_name}}.1_preprocess_data')


def main(filenames: List[str], preprocessing: Union[str, None], input_col: Union[str, int],
         sep: str = '{{default_sep}}', encoding: str = '{{default_encoding}}') -> None:
    '''Preprocesses some datasets

    Args:
        filenames (list<str>): Datasets filenames (actually paths relative to {{package_name}}-data)
        preprocessing (str): Preprocessing to be applied. All preprocessings are applied if None.
        input_col (str | int): Column to be preprocessed
    Kwargs:
        sep (str): Separator to use with the .csv files
        encoding (str): Encoding to use with the .csv files
    Raises:
        FileNotFoundError: If a given file does not exist in {{package_name}}-data
    '''
    logger.info("Data preprocessing ...")

    ##############################################
    # Manage preprocessing pipelines
    ##############################################

    # Get preprocess dictionnary
    preprocessors_dict = preprocess.get_preprocessors_dict()

    # Get preprocessing(s) to be applied
    if preprocessing is not None:
        # Check presence in preprocessors_dict
        if preprocessing not in preprocessors_dict.keys():
            raise ValueError(f"The given preprocessing {preprocessing} is not known.")
        preprocessing_list = [preprocessing]
    # By default, we apply every preprocessings
    else:
        preprocessing_list = list(preprocessors_dict.keys())


    ##############################################
    # Process each file, one by one
    ##############################################

    # Process each file, one by one
    for filename in filenames:

        # Get paths
        data_path = utils.get_data_path()
        dataset_path = os.path.join(data_path, filename)
        if not os.path.exists(dataset_path):
            raise FileNotFoundError(f"The file {dataset_path} does not exist")

        # For each preprocess: get the dataframe, apply the preprocessing & save
        for preprocess_str in preprocessing_list:

            # 'no_preprocess' must be ignored
            if preprocess_str == 'no_preprocess':
                continue
            gc.collect()  # Fix some OOM in case of huge datasets being preprocessed
            logger.info(f'Applying preprocessing {preprocess_str} on filename {filename}')

            # Get preprocessor
            preprocessor = preprocess.get_preprocessor(preprocess_str)
            # Get dataset
            df = pd.read_csv(dataset_path, sep=sep, encoding=encoding, dtype=str).fillna('')
            # Preprocess input_col
            x_col = utils.get_new_column_name(list(df.columns), 'preprocessed_text')
            df[x_col] = preprocessor(df[input_col])

            # Save preprocessed dataframe ({{default_encoding}}, '{{default_sep}}')
            # First line is a metadata with the name of the preprocess
            basename = Path(filename).stem
            dataset_preprocessed_path = os.path.join(data_path, f'{basename}_{preprocess_str}.csv')
            utils.to_csv(df, dataset_preprocessed_path, first_line=f'#{preprocess_str}', sep='{{default_sep}}', encoding='{{default_encoding}}')


if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument('-f', '--filenames', nargs='+', required=True, help='Datasets filenames (actually paths relative to {{package_name}}-data).')
    parser.add_argument('-p', '--preprocessing', default=None, help='Preprocessing to be applied. All preprocessings are applied if None.')
    parser.add_argument('--input_col', default=None, help='Column to be preprocessed')
    parser.add_argument('--sep', default='{{default_sep}}', help="Separator to use with the .csv files.")
    parser.add_argument('--encoding', default="{{default_encoding}}", help="Encoding to use with the .csv files.")
    args = parser.parse_args()
    main(filenames=args.filenames, preprocessing=args.preprocessing, input_col=args.input_col, sep=args.sep, encoding=args.encoding)



#!/usr/bin/env python3

## Finetune a pretrained transformer
#  /!\ NEEDS A LOT OF RESSOURCES /!\
# Copyright (C) <2018-2022>  <Agence Data Services, DSI Pôle Emploi>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# Ex: python 98_fine_tune_transformer.py -d dataset_train.csv -v dataset_valid.csv -p flaubert_small_cased -x text -o my_transformer -e 5 -b 64


import os
# Disable some tensorflow logs right away
os.environ['TF_CPP_MIN_LOG_LEVEL'] = '3'

import torch
import logging
import warnings
import argparse
import numpy as np
import pandas as pd
from typing import Union
from pathlib import Path
from transformers import CONFIG_NAME, WEIGHTS_NAME

from {{package_name}} import utils
from {{package_name}}.models_training import model_pytorch_language_model

# Disable some warnings
warnings.filterwarnings("ignore", category=FutureWarning)
warnings.filterwarnings("ignore", category=DeprecationWarning)

# Get logger
logger = logging.getLogger('{{package_name}}.98_fine_tune_transformer')



def main(dataset_train: str, dataset_valid: str, pretrained_model_name: str, x_col: Union[str, int],
         batch_size: int = 4, epochs: int = 1, output_dir: Union[str, None] = None,
         sep: str = '{{default_sep}}', encoding: str = '{{default_encoding}}') -> None:
    '''Finetunes a pretrained transformer

    Args:
        dataset_train (str): Name of the training dataset (actually a path relative to {{package_name}}-data)
        dataset_valid (str): Name of the validation dataset (actually a path relative to {{package_name}}-data)
        pretrained_model_name (str): Name of the tranformer model to be finetuned (actually a path relative to {{package_name}}-transformers)
        x_col (str | int): Name of the model's input column - x
    Kwargs:
        batch_size (int): Model's batch size
        epochs (int): Model's number of epochs
        output_dir (str): Name of the output model's directory (actually a path relative to {{package_name}}-data)
            If None, we just suffix the pretrained model name with '_finetuned'
        sep (str): Separator to use with the .csv files
        encoding (str): Encoding to use with the .csv files
    Raises:
        FileNotFoundError: If the training file does not exist in {{package_name}}-data
        FileNotFoundError: If the validation file does not exist in {{package_name}}-data
        FileNotFoundError: If the model does not exist in {{package_name}}-tranformers
        FileExistsError: If the output directory already exists
    '''
    logger.info("Finetuning a transformer model ...")

    # Manage output dir name
    if output_dir is None:
        output_dir = f"{pretrained_model_name}_finetuned"

    # Get data paths
    dataset_train = os.path.join(utils.get_data_path(), dataset_train)
    dataset_valid = os.path.join(utils.get_data_path(), dataset_valid)
    pretrained_model_path = os.path.join(utils.get_transformers_path(), pretrained_model_name)
    output_dir = os.path.join(utils.get_transformers_path(), output_dir)

    if not os.path.exists(dataset_train):
        raise FileNotFoundError(f"The file {dataset_train} does not exist")
    if not os.path.exists(dataset_valid):
        raise FileNotFoundError(f"The file {dataset_valid} does not exist")
    if not os.path.exists(pretrained_model_path):
        raise FileNotFoundError(f"The path {pretrained_model_path} does not exist")
    if not os.path.exists(output_dir):
        raise FileExistsError(f"Path {output_dir} already exists.")

    # Load data
    df_train = pd.read_csv(dataset_train, sep=sep, encoding=encoding)
    df_valid = pd.read_csv(dataset_valid, sep=sep, encoding=encoding)
    # Legacy: add a fake target column to comply with ModelPytorch
    y_col = ["dummy"]
    df_train[y_col] = pd.Series(np.random.randint(0, 2, df_train.shape[0]))
    df_valid[y_col] = pd.Series(np.random.randint(0, 2, df_valid.shape[0]))
    x_train, y_train = df_train[x_col], df_train[y_col]
    x_valid, y_valid = df_valid[x_col], df_valid[y_col]

    # Create language model
    model_pytorch_language_model.TORCH_DEVICE
    model = model_pytorch_language_model.ModelPyTorchLanguageModel(pretrained_model_name, multi_label=True, batch_size=batch_size, epochs=epochs)

    # Train it
    model.fit(x_train, y_train, x_valid=x_valid, y_valid=y_valid, with_shuffle=True)
    # Add a single forward
    model.forward()

    # Save
    os.makedirs(output_dir)  # We already cheked it does not exist
    model_to_save = model.model.model.module if hasattr(model.model.model, "module") else model.model.model
    output_model_file = os.path.join(output_dir, WEIGHTS_NAME)
    output_config_file = os.path.join(output_dir, CONFIG_NAME)
    torch.save(model_to_save.state_dict(), output_model_file)
    model_to_save.config.to_json_file(output_config_file)
    model.tokenizer.save_vocabulary(output_dir)


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument('-d', '--dataset_train', required=True, help="Name of the training dataset (actually a path relative to {{package_name}}-data)")
    parser.add_argument('-v', '--dataset_valid', required=True, help="Name of the validation dataset (actually a path relative to {{package_name}}-data)")
    parser.add_argument('-p', '--pretrained_model_name', required=True, help="Name of the tranformer model to be finetuned (actually a path relative to {{package_name}}-transformers)")
    parser.add_argument('-x', '--x_col', required=True, help="Name of the model's input column - x")
    parser.add_argument('-o', '--output_dir', default=None, help="Model's batch size")
    parser.add_argument('-e', '--epochs', default=1, type=int, help="Model's number of epochs")
    parser.add_argument('-b', '--batch_size', default=4, type=int, help="Name of the output model's directory (actually a path relative to {{package_name}}-data)")
    parser.add_argument('--sep', default='{{default_sep}}', help="Separator to use with the .csv files")
    parser.add_argument('--encoding', default="{{default_encoding}}", help="Encoding to use with the .csv files")
    args = parser.parse_args()
    main(output_dir=args.output_dir, pretrained_model_name=args.pretrained_model_name,
         dataset_train=args.dataset_train, dataset_valid=args.dataset_valid,
         x_col = args.x_col, batch_size=args.batch_size, epochs=args.epochs,
         sep=args.sep, encoding=args.encoding)
